@@ -11,6 +11,9 @@
 
 #ifndef OPD_EVENTS_H
 
+#include "op_types.h"
+
+#include <stdint.h>
 #include <time.h>
 
 /** event description for setup (perfmon) and mangling */
@@ -37,6 +40,8 @@ struct opd_header;
 
 /** fill the sample file header with event info etc. */
 void fill_header(struct opd_header * header, unsigned long counter,
-                 int is_kernel, int cg_to_is_kernel, time_t mtime);
+		 vma_t anon_start, vma_t anon_end,
+		 int is_kernel, int cg_to_is_kernel,
+                 int spu_samples, uint64_t embed_offset, time_t mtime);
 
 #endif /* OPD_EVENTS_H */
