@@ -20,28 +20,44 @@
 #include <fstream>
 #include <algorithm>
 
+#if QT3_SUPPORT
+#include <Qt/qlineedit.h>
+#include <Qt/qcheckbox.h>
+#include <Qt/qtabwidget.h>
+#include <Qt/qmessagebox.h>
+#include <Qt/qvalidator.h>
+#include <Qt/qlabel.h>
+#include <Qt/qpushbutton.h>
+#include <Qt/q3listview.h>
+#include <Qt/q3combobox.h>
+#include <Qt/q3listbox.h>
+#include <Qt/q3filedialog.h>
+#include <Qt/q3buttongroup.h>
+#include <Qt/q3header.h>
+#else
 #include <qlineedit.h>
-#include <qlistview.h>
-#include <qcombobox.h>
-#include <qlistbox.h>
-#include <qfiledialog.h>
-#include <qbuttongroup.h>
 #include <qcheckbox.h>
 #include <qtabwidget.h>
 #include <qmessagebox.h>
 #include <qvalidator.h>
 #include <qlabel.h>
 #include <qpushbutton.h>
+#include <qlistview.h>
+#include <qcombobox.h>
+#include <qlistbox.h>
+#include <qfiledialog.h>
+#include <qbuttongroup.h>
 #include <qheader.h>
+#define Q3ListView QListView
+#endif
 
-#include "config.h"
 #include "oprof_start.h"
 #include "op_config.h"
-#include "op_config_24.h"
 #include "string_manip.h"
 #include "op_cpufreq.h"
 #include "op_alloc_counter.h"
 #include "oprof_start_util.h"
+#include "file_manip.h"
 
 #include "op_hw_config.h"
 
@@ -91,6 +107,8 @@ static char const * red_xpm[] = {
 
 static QPixmap * green_pixmap;
 static QPixmap * red_pixmap;
+static op_event_descr null_evt;
+
 
 
 op_event_descr::op_event_descr()
@@ -139,9 +157,17 @@ oprof_start::oprof_start()
 	if (is_26) {
 		note_table_size_edit->hide();
 		note_table_size_label->hide();
-		// FIXME: can adapt to 2.6 ...
-		buffer_size_edit->hide();
-		buffer_size_label->hide();
+		if (!op_file_readable("/dev/oprofile/backtrace_depth")) {
+			callgraph_depth_label->hide();
+			callgraph_depth_edit->hide();
+		}
+	} else {
+		callgraph_depth_label->hide();
+		callgraph_depth_edit->hide();
+		buffer_watershed_label->hide();
+		buffer_watershed_edit->hide();
+		cpu_buffer_size_label->hide();
+		cpu_buffer_size_edit->hide();
 	}
 
 	// setup the configuration page.
@@ -150,7 +176,10 @@ oprof_start::oprof_start()
 	no_vmlinux->setChecked(config.no_kernel);
 
 	buffer_size_edit->setText(QString().setNum(config.buffer_size));
+	buffer_watershed_edit->setText(QString().setNum(config.buffer_watershed));
+	cpu_buffer_size_edit->setText(QString().setNum(config.cpu_buffer_size));
 	note_table_size_edit->setText(QString().setNum(config.note_table_size));
+	callgraph_depth_edit->setText(QString().setNum(config.callgraph_depth));
 	verbose->setChecked(config.verbose);
 	separate_lib_cb->setChecked(config.separate_lib);
 	separate_kernel_cb->setChecked(config.separate_kernel);
@@ -164,8 +193,13 @@ oprof_start::oprof_start()
 	QIntValidator * iv;
 	iv = new QIntValidator(OP_MIN_BUF_SIZE, OP_MAX_BUF_SIZE, buffer_size_edit);
 	buffer_size_edit->setValidator(iv);
-	iv = new QIntValidator(OP_MIN_NOTE_TABLE_SIZE, OP_MAX_NOTE_TABLE_SIZE, note_table_size_edit);
 	note_table_size_edit->setValidator(iv);
+	iv = new QIntValidator(0, INT_MAX, callgraph_depth_edit);
+	callgraph_depth_edit->setValidator(iv);
+	iv = new QIntValidator(0, INT_MAX, buffer_watershed_edit);
+	buffer_watershed_edit->setValidator(iv);
+	iv = new QIntValidator(0, OP_MAX_CPU_BUF_SIZE, cpu_buffer_size_edit);
+	cpu_buffer_size_edit->setValidator(iv);
 
 	// daemon status timer
 	startTimer(5000);
@@ -250,14 +284,13 @@ void oprof_start::fill_events()
 namespace {
 
 /// find the first item with the given text in column 0 or return NULL
-QListViewItem * findItem(QListView * view, char const * name)
+Q3ListViewItem * findItem(Q3ListView * view, char const * name)
 {
 	// Qt 2.3.1 does not have QListView::findItem()
-	QListViewItem * item = view->firstChild();
+	Q3ListViewItem * item = view->firstChild();
 
-	while (item && strcmp(item->text(0).latin1(), name)) {
+	while (item && strcmp(item->text(0).latin1(), name))
 		item = item->nextSibling();
-	}
 
 	return item;
 }
@@ -275,7 +308,7 @@ void oprof_start::setup_default_event()
 	event_cfgs[descr.name].user_ring_count = 1;
 	event_cfgs[descr.name].os_ring_count = 1;
 
-	QListViewItem * item = findItem(events_list, descr.name);
+	Q3ListViewItem * item = findItem(events_list, descr.name);
 	if (item)
 		item->setSelected(true);
 }
@@ -283,7 +316,7 @@ void oprof_start::setup_default_event()
 
 void oprof_start::read_set_events()
 {
-	string name = get_user_filename(".oprofile/daemonrc");
+	string name = get_config_filename(".oprofile/daemonrc");
 
 	ifstream in(name.c_str());
 
@@ -300,12 +333,12 @@ void oprof_start::read_set_events()
 		string const val = split(str, '=');
 		string const name = str;
 
-		if (!is_prefix(name, "CHOSEN_EVENTS["))
+		if (!is_prefix(name, "CHOSEN_EVENTS_"))
 			continue;
 
 		one_enabled = true;
 
-		// CHOSEN_EVENTS[0]=CPU_CLK_UNHALTED:10000:0:1:1
+		// CHOSEN_EVENTS_#nr=CPU_CLK_UNHALTED:10000:0:1:1
 		vector<string> parts = separate_token(val, ':');
 
 		if (parts.size() != 5 && parts.size() != 2) {
@@ -332,7 +365,7 @@ void oprof_start::read_set_events()
 			event_cfgs[ev_name].os_ring_count = 1;
 		}
 
-		QListViewItem * item = findItem(events_list, ev_name.c_str());
+		Q3ListViewItem * item = findItem(events_list, ev_name.c_str());
 		if (item)
 			item->setSelected(true);
 	}
@@ -345,7 +378,7 @@ void oprof_start::read_set_events()
 
 void oprof_start::load_config_file()
 {
-	string name = get_user_filename(".oprofile/daemonrc");
+	string name = get_config_filename(".oprofile/daemonrc");
 
 	ifstream in(name.c_str());
 	if (!in) {
@@ -391,6 +424,7 @@ void oprof_start::timerEvent(QTimerEvent *)
 	flush_profiler_data_btn->setEnabled(dstat.running);
 	stop_profiler_btn->setEnabled(dstat.running);
 	start_profiler_btn->setEnabled(!dstat.running);
+	reset_sample_files_btn->setEnabled(!dstat.running);
 
 	if (!dstat.running) {
 		daemon_label->setText("Profiler is not running.");
@@ -417,8 +451,8 @@ void oprof_start::fill_events_listbox()
 	setUpdatesEnabled(false);
 
 	for (vector<op_event_descr>::reverse_iterator cit = v_events.rbegin();
-		cit != v_events.rend(); ++cit) {
-		new QListViewItem(events_list, cit->name.c_str());
+	     cit != v_events.rend(); ++cit) {
+		new Q3ListViewItem(events_list, cit->name.c_str());
 	}
 
 	setUpdatesEnabled(true);
@@ -449,7 +483,7 @@ void oprof_start::display_event(op_event_descr const & descr)
 }
 
 
-bool oprof_start::is_selectable_event(QListViewItem * item)
+bool oprof_start::is_selectable_event(Q3ListViewItem * item)
 {
 	if (item->isSelected())
 		return true;
@@ -468,7 +502,7 @@ bool oprof_start::is_selectable_event(QListViewItem * item)
 
 void oprof_start::draw_event_list()
 {
-	QListViewItem * cur;
+	Q3ListViewItem * cur;
 	for (cur = events_list->firstChild(); cur; cur = cur->nextSibling()) {
 		if (is_selectable_event(cur))
 			cur->setPixmap(0, *green_pixmap);
@@ -482,10 +516,9 @@ bool oprof_start::alloc_selected_events() const
 {
 	vector<op_event const *> events;
 
-	set<QListViewItem *>::const_iterator it;
-	for (it = selected_events.begin(); it != selected_events.end(); ++it) {
-		events.push_back(find_event_by_name((*it)->text(0).latin1()));
-	}
+	set<Q3ListViewItem *>::const_iterator it;
+	for (it = selected_events.begin(); it != selected_events.end(); ++it)
+		events.push_back(find_event_by_name((*it)->text(0).latin1(),0,0));
 
 	size_t * map =
 		map_event_to_counter(&events[0], events.size(), cpu_type);
@@ -503,26 +536,24 @@ void oprof_start::event_selected()
 	// (de)selected item so we record a set of selected items and diff
 	// it in the appropriate way with the previous list of selected items.
 
-	set<QListViewItem *> current_selection;
-	QListViewItem * cur;
+	set<Q3ListViewItem *> current_selection;
+	Q3ListViewItem * cur;
 	for (cur = events_list->firstChild(); cur; cur = cur->nextSibling()) {
-		if (cur->isSelected()) {
+		if (cur->isSelected())
 			current_selection.insert(cur);
-		}
 	}
 
 	// First remove the deselected item.
-	vector<QListViewItem *> new_deselected;
+	vector<Q3ListViewItem *> new_deselected;
 	set_difference(selected_events.begin(), selected_events.end(),
 		       current_selection.begin(), current_selection.end(),
 		       back_inserter(new_deselected));
-	vector<QListViewItem *>::const_iterator it;
-	for (it = new_deselected.begin(); it != new_deselected.end(); ++it) {
+	vector<Q3ListViewItem *>::const_iterator it;
+	for (it = new_deselected.begin(); it != new_deselected.end(); ++it)
 		selected_events.erase(*it);
-	}
 
 	// Now try to add the newly selected item if enough HW resource exists
-	vector<QListViewItem *> new_selected;
+	vector<Q3ListViewItem *> new_selected;
 	set_difference(current_selection.begin(), current_selection.end(),
 		       selected_events.begin(), selected_events.end(),
 		       back_inserter(new_selected));
@@ -543,7 +574,7 @@ void oprof_start::event_selected()
 }
 
 
-void oprof_start::event_over(QListViewItem * item)
+void oprof_start::event_over(Q3ListViewItem * item)
 {
 	op_event_descr const & descr = locate_event(item->text(0).latin1());
 
@@ -551,10 +582,10 @@ void oprof_start::event_over(QListViewItem * item)
 	if (!is_selectable_event(item)) {
 		help_str += " conflicts with:";
 
-		set<QListViewItem *>::const_iterator it;
+		set<Q3ListViewItem *>::const_iterator it;
 		for (it = selected_events.begin(); 
 		     it != selected_events.end(); ) {
-			QListViewItem * temp = *it;
+			Q3ListViewItem * temp = *it;
 			selected_events.erase(it++);
 			if (is_selectable_event(item)) {
 				help_str += " ";
@@ -605,8 +636,7 @@ bool oprof_start::record_config()
 	config.kernel_filename = kernel_filename_edit->text().latin1();
 	config.no_kernel = no_vmlinux->isChecked();
 
-	QString const t = buffer_size_edit->text();
-	uint temp = t.toUInt();
+	uint temp = buffer_size_edit->text().toUInt();
 	if (temp < OP_MIN_BUF_SIZE || temp > OP_MAX_BUF_SIZE) {
 		ostringstream error;
 
@@ -620,19 +650,55 @@ bool oprof_start::record_config()
 	}
 	config.buffer_size = temp;
 
-	temp = note_table_size_edit->text().toUInt();
-	if (temp < OP_MIN_NOTE_TABLE_SIZE || temp > OP_MAX_NOTE_TABLE_SIZE) {
+	temp = buffer_watershed_edit->text().toUInt();
+	// watershed above half of buffer size make little sense.
+	if (temp > config.buffer_size / 2) {
 		ostringstream error;
 
-		error << "note table size out of range: " << temp
-		      << " valid range is [" << OP_MIN_NOTE_TABLE_SIZE << ", "
-		      << OP_MAX_NOTE_TABLE_SIZE << "]";
+		error << "buffer watershed out of range: " << temp
+		      << " valid range is [0 (use default), buffer size/2] "
+		      << "generally 0.25 * buffer size is fine";
 
 		QMessageBox::warning(this, 0, error.str().c_str());
 
 		return false;
 	}
-	config.note_table_size = temp;
+	config.buffer_watershed = temp;
+
+	temp = cpu_buffer_size_edit->text().toUInt();
+	if ((temp != 0 && temp < OP_MIN_CPU_BUF_SIZE) ||
+	    temp > OP_MAX_CPU_BUF_SIZE) {
+		ostringstream error;
+
+		error << "cpu buffer size out of range: " << temp
+		      << " valid range is [" << OP_MIN_CPU_BUF_SIZE << ", "
+		      << OP_MAX_CPU_BUF_SIZE << "] (size = 0: use default)";
+
+		QMessageBox::warning(this, 0, error.str().c_str());
+
+		return false;
+	}
+	config.cpu_buffer_size = temp;
+
+	// Note table size is a 2.4 kernel thingy.  We'll just set it to
+	// zero here since it's not used.
+	// FIXME: Some day we should do a proper cleanup of the GUI to remove
+	//        all leftover junk having anything to do with 2.4 kernels.
+	config.note_table_size = 0;
+
+	temp = callgraph_depth_edit->text().toUInt();
+	if (temp > INT_MAX) {
+		ostringstream error;
+
+		error << "callgraph depth  out of range: " << temp
+		      << " valid range is [" << 0 << ", "
+		      << INT_MAX << "]";
+
+		QMessageBox::warning(this, 0, error.str().c_str());
+
+		return false;
+	}
+	config.callgraph_depth = temp;
 
 	config.verbose = verbose->isChecked();
 	config.separate_lib = separate_lib_cb->isChecked();
@@ -716,6 +782,7 @@ void oprof_start::hide_masks()
 
 void oprof_start::setup_unit_masks(op_event_descr const & descr)
 {
+#define OP_MAX_HANDLED_UMS 16
 	op_unit_mask const * um = descr.unit;
 
 	hide_masks();
@@ -723,12 +790,23 @@ void oprof_start::setup_unit_masks(op_event_descr const & descr)
 	if (!um || um->unit_type_mask == utm_mandatory)
 		return;
 
+	if (um->num > OP_MAX_HANDLED_UMS) {
+		ostringstream error;
+
+		error << "Number of unit masks (" << um->num << ") is greater than max allowed ("
+		      << OP_MAX_HANDLED_UMS << ").";
+		QMessageBox::warning(this, 0, error.str().c_str());
+		return;
+	}
+
+
 	event_setting & cfg = event_cfgs[descr.name];
 
 	unit_mask_group->setExclusive(um->unit_type_mask == utm_exclusive);
 
-	for (size_t i = 0; i < um->num ; ++i) {
-		QCheckBox * check = 0;
+	QCheckBox * check = NULL;
+
+	for (size_t i = 0; i < um->num; ++i) {
 		switch (i) {
 			case 0: check = check0; break;
 			case 1: check = check1; break;
@@ -748,22 +826,13 @@ void oprof_start::setup_unit_masks(op_event_descr const & descr)
 			case 15: check = check15; break;
 		}
 		check->setText(um->um[i].desc);
-		if (um->unit_type_mask == utm_exclusive) {
+		if (um->unit_type_mask == utm_exclusive)
 			check->setChecked(cfg.umask == um->um[i].value);
-		} else {
-			// The last descriptor contains a mask that enable all
-			// value so we must enable the last check box only if
-			// all bits are on.
-			if (i == um->num - 1) {
-				check->setChecked(cfg.umask == um->um[i].value);
-			} else {
-				check->setChecked(cfg.umask & um->um[i].value);
-			}
-		}
+		else
+			check->setChecked(cfg.umask & um->um[i].value);
+
 		check->show();
 	}
-	unit_mask_group->setMinimumSize(unit_mask_group->sizeHint());
-	setup_config_tab->setMinimumSize(setup_config_tab->sizeHint());
 }
 
 
@@ -793,7 +862,7 @@ void oprof_start::on_start_profiler()
 
 	bool one_enable = false;
 
-	QListViewItem * cur;
+	Q3ListViewItem * cur;
 	for (cur = events_list->firstChild(); cur; cur = cur->nextSibling()) {
 		if (!cur->isSelected())
 			continue;
@@ -832,7 +901,7 @@ void oprof_start::on_start_profiler()
 		    cfg.umask == 0) {
 			ostringstream out;
 
-			out << "event " << descr.name<< " invalid unit mask: "
+			out << "event " << descr.name << " invalid unit mask: "
 			    << cfg.umask << endl;
 
 			QMessageBox::warning(this, 0, out.str().c_str());
@@ -895,7 +964,7 @@ bool oprof_start::save_config()
 	vector<string> tmpargs;
 	tmpargs.push_back("--setup");
 
-	QListViewItem * cur;
+	Q3ListViewItem * cur;
 	for (cur = events_list->firstChild(); cur; cur = cur->nextSibling()) {
 		if (!cur->isSelected())
 			continue;
@@ -926,11 +995,16 @@ bool oprof_start::save_config()
 		args.push_back("--vmlinux=" + config.kernel_filename);
 	}
 
-	if (op_get_interface() == OP_INTERFACE_24) {
-		args.push_back("--buffer-size=" +
-		       op_lexical_cast<string>(config.buffer_size));
-		args.push_back("--note-table-size=" +
-		       op_lexical_cast<string>(config.note_table_size));
+	args.push_back("--buffer-size=" +
+	               op_lexical_cast<string>(config.buffer_size));
+
+	args.push_back("--buffer-watershed=" +
+	               op_lexical_cast<string>(config.buffer_watershed));
+	args.push_back("--cpu-buffer-size=" +
+	               op_lexical_cast<string>(config.cpu_buffer_size));
+	if (op_file_readable("/dev/oprofile/backtrace_depth")) {
+		args.push_back("--callgraph=" +
+		               op_lexical_cast<string>(config.callgraph_depth));
 	}
 
 	string sep = "--separate=";
@@ -975,6 +1049,22 @@ void oprof_start::on_separate_kernel_cb_changed(int state)
 		separate_lib_cb->setChecked(true);
 }
 
+void oprof_start::on_reset_sample_files()
+{
+	int ret = QMessageBox::warning(this, 0, "Are you sure you want to "
+	       "reset your last profile session ?", "Yes", "No", 0, 0, 1);
+	if (!ret) {
+		vector<string> args;
+		args.push_back("--reset");
+		if (!do_exec_command(OP_BINDIR "/opcontrol", args))
+			// the next timer event will overwrite the message
+			daemon_label->setText("Last profile session reseted.");
+		else
+			QMessageBox::warning(this, 0,
+			     "Can't reset profiling session.");
+	}
+}
+
 
 /// function object for matching against name
 class event_name_eq {
@@ -990,5 +1080,15 @@ public:
 // helper to retrieve an event descr through its name.
 op_event_descr const & oprof_start::locate_event(string const & name) const
 {
-	return *(find_if(v_events.begin(), v_events.end(), event_name_eq(name)));
+	vector<op_event_descr>::const_iterator it = find_if(v_events.begin(), v_events.end(),
+	                                                    event_name_eq(name));
+
+	// Failure to find the event should NEVER happen; I had to put this here to silence Coverity.
+	if (it != v_events.end())
+		return *it;
+
+	QMessageBox::warning((oprof_start *)this, 0, "Could not locate event. Returning null event.");
+	null_evt.help_str = "";
+	null_evt.name = "";
+	return null_evt;
 }
