@@ -11,11 +11,17 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <sys/utsname.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fnmatch.h>
+#include <elf.h>
+#include <link.h>
 
 #include "op_cpu_type.h"
 #include "op_hw_specific.h"
@@ -107,6 +113,11 @@ static struct cpu_descr const cpu_descrs[MAX_CPU_TYPE] = {
 	{ "ARM Cortex-A5", "arm/armv7-ca5", CPU_ARM_V7_CA5, 3 },
 	{ "ARM Cortex-A7", "arm/armv7-ca7", CPU_ARM_V7_CA7, 5 },
 	{ "ARM Cortex-A15", "arm/armv7-ca15", CPU_ARM_V7_CA15, 7 },
+	{ "Intel Haswell microarchitecture", "i386/haswell", CPU_HASWELL, 4 },
+	{ "IBM zEnterprise EC12", "s390/zEC12", CPU_S390_ZEC12, 1 },
+	{ "AMD64 generic", "x86-64/generic", CPU_AMD64_GENERIC, 4 },
+	{ "IBM Power Architected Events V1", "ppc64/architected_events_v1", CPU_PPC64_ARCH_V1, 6 },
+	{ "ppc64 POWER8", "ppc64/power8", CPU_PPC64_POWER8, 6 },
 };
  
 static size_t const nr_cpu_descrs = sizeof(cpu_descrs) / sizeof(struct cpu_descr);
@@ -165,11 +176,127 @@ static char * _get_cpuinfo_cpu_type(char * buf, int len, const char * prefix)
 	return _get_cpuinfo_cpu_type_line(buf, len, prefix, 1);
 }
 
+// The aux vector stuff below is currently only used by ppc64 arch
+static ElfW(auxv_t) * auxv_buf = NULL;
+
+static ElfW(auxv_t) * _auxv_fetch()
+{
+	ElfW(auxv_t) * auxv_temp = (ElfW(auxv_t) *)auxv_buf;
+	int auxv_f;
+	size_t page_size = getpagesize();
+	ssize_t bytes;
+
+
+	if(auxv_temp == NULL) {
+		auxv_f = open("/proc/self/auxv", O_RDONLY);
+
+		if(auxv_f == -1) {
+			perror("Cannot open /proc/self/auxv");
+			fprintf(stderr, "Assuming native platform profiling is supported.\n");
+			return NULL;
+		}
+		else {
+			auxv_temp = (ElfW(auxv_t) *)malloc(page_size);
+			if (!auxv_temp) {
+				perror("Allocation of space for auxv failed.");
+				close(auxv_f);
+				return NULL;
+			}
+			bytes = read(auxv_f, (void *)auxv_temp, page_size);
+
+			if (bytes <= 0) {
+				free(auxv_temp);
+				close(auxv_f);
+				perror("Error /proc/self/auxv read failed");
+				return NULL;
+			}
+
+			if (close(auxv_f)) {
+				perror("Error close failed");
+				fprintf(stderr, "Recoverable error. Continuing.\n");
+			}
+		}
+		auxv_buf = auxv_temp;
+	}
+	return (ElfW(auxv_t) *)auxv_temp;
+}
+
+
+static const char * fetch_at_hw_platform(ElfW(Addr) type)
+{
+	int i = 0;
+	const char * platform = NULL;
+	ElfW(auxv_t) * my_auxv = NULL;
+
+	if ((my_auxv = (ElfW(auxv_t)*) _auxv_fetch()) == NULL)
+		return NULL;
+
+	do {
+		if(my_auxv[i].a_type == type) {
+			platform = (const char *)my_auxv[i].a_un.a_val;
+			break;
+		}
+		i++;
+	} while (my_auxv[i].a_type != AT_NULL);
+
+	return platform;
+}
+
+static void release_at_hw_platform(void)
+{
+	if (auxv_buf) {
+		free(auxv_buf);
+		auxv_buf = NULL;
+	}
+}
+
+static op_cpu _try_ppc64_arch_generic_cpu(void)
+{
+	const char * platform, * base_platform;
+	op_cpu cpu_type = CPU_NO_GOOD;
+
+	platform = fetch_at_hw_platform(AT_PLATFORM);
+	base_platform = fetch_at_hw_platform(AT_BASE_PLATFORM);
+	if (!platform || !base_platform) {
+		fprintf(stderr, "NULL returned for one or both of AT_PLATFORM/AT_BASE_PLATFORM\n");
+		fprintf(stderr, "AT_PLATFORM: %s; \tAT_BASE_PLATFORM: %s\n", platform, base_platform);
+		release_at_hw_platform();
+		return cpu_type;
+	}
+	// FIXME whenever a new IBM Power processor is added -- need to ensure
+	// we're returning the correct version of the architected events file.
+	if (strcmp(platform, base_platform)) {
+		// If platform and base_platform differ by only a "+" at the end of the name, we
+		// consider these equal.
+		int platforms_are_equivalent = 0;
+		size_t p1_len, p2_len;
+		p1_len = strlen(platform);
+		p2_len = strlen(base_platform);
+		if (p2_len == (p1_len + 1)) {
+			if ((strncmp(platform, base_platform, p1_len) == 0) &&
+					(base_platform[p2_len - 1] == '+')) {
+				platforms_are_equivalent = 1;
+			}
+		}
+		if (!platforms_are_equivalent) {
+			if (strcmp(platform, "power7") == 0)
+				cpu_type = CPU_PPC64_ARCH_V1;
+		}
+	}
+	release_at_hw_platform();
+	return cpu_type;
+}
+
 static op_cpu _get_ppc64_cpu_type(void)
 {
 	int i;
 	size_t len;
 	char line[100], cpu_type_str[64], cpu_name_lowercase[64], * cpu_name;
+	op_cpu cpu_type = CPU_NO_GOOD;
+
+	cpu_type = _try_ppc64_arch_generic_cpu();
+	if (cpu_type != CPU_NO_GOOD)
+		return cpu_type;
 
 	cpu_name = _get_cpuinfo_cpu_type(line, 100, "cpu");
 	if (!cpu_name)
@@ -182,7 +309,8 @@ static op_cpu _get_ppc64_cpu_type(void)
 	cpu_type_str[0] = '\0';
 	strcat(cpu_type_str, "ppc64/");
 	strncat(cpu_type_str, cpu_name_lowercase, len);
-	return op_get_cpu_number(cpu_type_str);
+	cpu_type = op_get_cpu_number(cpu_type_str);
+	return cpu_type;
 }
 
 static op_cpu _get_arm_cpu_type(void)
@@ -343,12 +471,16 @@ static op_cpu _get_intel_cpu_type(void)
 
 static op_cpu _get_amd_cpu_type(void)
 {
-	unsigned eax, family, model;
+	unsigned eax, family;
 	op_cpu ret = CPU_NO_GOOD;
+	char buf[20] = {'\0'};
 
 	eax = cpuid_signature();
 	family = cpu_family(eax);
-	model = cpu_model(eax);
+
+	/* These family does not exist in the past.*/
+	if (family < 0x0f || family == 0x13)
+		return ret;
 
 	switch (family) {
 	case 0x0f:
@@ -358,24 +490,17 @@ static op_cpu _get_amd_cpu_type(void)
 		ret = op_get_cpu_number("x86-64/family10");
 		break;
 	case 0x11:
-		ret = op_get_cpu_number("x86-64/family11h");
-		break;
 	case 0x12:
-		ret = op_get_cpu_number("x86-64/family12h");
-		break;
 	case 0x14:
-		ret = op_get_cpu_number("x86-64/family14h");
-		break;
 	case 0x15:
-		switch (model) {
-		case 0x00 ... 0x0f:
-			ret = op_get_cpu_number("x86-64/family15h");
-			break;		
-		default:
-			break;
-		}
+		/* From family11h and forward, we use the same naming scheme */
+		snprintf(buf, 20, "x86-64/family%xh", family);
+		ret = op_get_cpu_number(buf);
 		break;
 	default:
+		/* Future processors */
+		snprintf(buf, 20, "x86-64/generic");
+		ret = op_get_cpu_number(buf);
 		break;
 	}
 
@@ -501,6 +626,7 @@ op_cpu op_cpu_base_type(op_cpu cpu_type)
 	case CPU_CORE_I7:
 	case CPU_ATOM:
 	case CPU_NEHALEM:
+	case CPU_HASWELL:
 	case CPU_WESTMERE:
 	case CPU_SANDYBRIDGE:
 	case CPU_IVYBRIDGE:
