@@ -114,7 +114,8 @@ unsigned parse_extra(const char *s)
 	unsigned v, w;
 	int o;
 
-	v = 0;
+	/* This signifies that the first word of the description is unique */
+	v = EXTRA_NONE;
 	while (*s) {
 		if (isspace(*s))
 			break;
@@ -130,6 +131,9 @@ unsigned parse_extra(const char *s)
 		} else if (strisprefix(s, "any")) {
 			v |= EXTRA_ANY;
 			s += 3;
+		} else if (strisprefix(s, "pebs")) {
+			v |= EXTRA_PEBS;
+			s += 4;
 		} else {
 			parse_error("Illegal extra field modifier");
 		}
@@ -193,7 +197,12 @@ static void parse_um(struct op_unit_mask * um, char const * line)
 			if (seen_default)
 				parse_error("duplicate default: tag");
 			seen_default = 1;
-			um->default_mask = parse_hex(tagend);
+			if (0 != strncmp(tagend, "0x", 2)) {
+				um->default_mask_name = op_xstrndup(
+					tagend, valueend - tagend);
+			} else {
+				um->default_mask = parse_hex(tagend);
+			}
 		} else {
 			parse_error("invalid unit mask tag");
 		}
@@ -211,32 +220,50 @@ static void parse_um(struct op_unit_mask * um, char const * line)
 
 
 /* \t0x08 (M)odified cache state */
-/* \t0x08 extra:inv,cmask=... (M)odified cache state */
+/* \t0x08 extra:inv,cmask=... mod_cach_state (M)odified cache state */
 static void parse_um_entry(struct op_described_um * entry, char const * line)
 {
 	char const * c = line;
 
+	/* value */
 	c = skip_ws(c);
 	entry->value = parse_hex(c);
-	c = skip_nonws(c);
 
+	/* extra: */
+	c = skip_nonws(c);
 	c = skip_ws(c);
+	if (!*c)
+		goto invalid_out;
+
 	if (strisprefix(c, "extra:")) {
 		c += 6;
 		entry->extra = parse_extra(c);
+		/* named mask */
 		c = skip_nonws(c);
-	} else
+		c = skip_ws(c);
+		if (!*c)
+			goto invalid_out;
+
+		/* "extra:" !!ALWAYS!! followed by named mask */
+		entry->name = op_xstrndup(c, strcspn(c, " \t"));
+		c = skip_nonws(c);
+		c = skip_ws(c);
+	} else {
 		entry->extra = 0;
+	}
 
-	if (!*c)
-		parse_error("invalid unit mask entry");
+	/* desc */
+	if (!*c) {
+		/* This is a corner case where the named unit mask entry
+		 * only has one word.  This should really be fixed in the
+		 * unit_mask file */
+		entry->desc = xstrdup(entry->name);
+	} else
+		entry->desc = xstrdup(c);
+	return;
 
-	c = skip_ws(c);
-
-	if (!*c)
-		parse_error("invalid unit mask entry");
-
-	entry->desc = xstrdup(c);
+invalid_out:
+	parse_error("invalid unit mask entry");
 }
 
 
@@ -253,6 +280,7 @@ static void free_unit_mask(struct op_unit_mask * um)
 {
 	list_del(&um->um_next);
 	free(um);
+	um = NULL;
 }
 
 /*
@@ -286,8 +314,6 @@ static void read_unit_masks(char const * file)
 		} else {
 			if (!um)
 				parse_error("no unit mask name line");
-			if (um->num >= MAX_UNIT_MASK)
-				parse_error("oprofile: maximum unit mask entries exceeded");
 
 			parse_um_entry(&um->um[um->num], line);
 			++(um->num);
@@ -383,7 +409,7 @@ static struct op_unit_mask * merge_um(char * value)
 	if (type == -1U)
 		parse_error("Empty unit mask");
 	new->unit_type_mask = type;
-	return new;		
+	return new;
 }
 
 /* parse either a "tag:value" or a ": trailing description string" */
@@ -497,7 +523,7 @@ static void read_events(char const * file)
 		event = new_event();
 		event->filter = -1;
 		event->ext = NULL;
-		
+
 		c = line;
 		while (next_token(&c, &name, &value)) {
 			if (strcmp(name, "name") == 0) {
@@ -558,6 +584,7 @@ static void read_events(char const * file)
 				c = skip_ws(c);
 				if (*c != '\0' && *c != '#')
 					parse_error("non whitespace after include:");
+				break;
 			} else {
 				parse_error("unknown tag");
 			}
@@ -601,10 +628,18 @@ static int check_unit_mask(struct op_unit_mask const * um,
 				"(%s)\n", um->name, cpu_name);
 			err = EXIT_FAILURE;
 		}
-	} else {
-		for (i = 0; i < um->num; ++i) {
-			if (um->default_mask == um->um[i].value)
-				break;
+	} else if (um->unit_type_mask == utm_exclusive) {
+		if (um->default_mask_name) {
+			for (i = 0; i < um->num; ++i) {
+				if (0 == strcmp(um->default_mask_name,
+						um->um[i].name))
+					break;
+			}
+		} else {
+			for (i = 0; i < um->num; ++i) {
+				if (um->default_mask == um->um[i].value)
+					break;
+			}
 		}
 
 		if (i == um->num) {
@@ -639,7 +674,7 @@ static void load_events_name(const char *cpu_name)
 
 	read_unit_masks(um_file);
 	read_events(event_file);
-	
+
 	free(um_file);
 	free(event_file);
 }
@@ -704,6 +739,7 @@ static void load_events(op_cpu cpu_type)
 	unit_mask->unit_type_mask = utm_mandatory;
 	unit_mask->um[0].extra = 0;
 	unit_mask->um[0].value = 0;
+	unit_mask->um[0].name = xstrdup("");
 	unit_mask->um[0].desc = xstrdup("No unit mask");
 	unit_mask->used = 1;
 
@@ -922,13 +958,14 @@ char const * find_mapping_for_event(u32 nr, op_cpu cpu_type)
 		case CPU_PPC64_POWER6:
 		case CPU_PPC64_POWER7:
 		case CPU_PPC64_IBM_COMPAT_V1:
+		// For ppc64 types of CPU_PPC64_ARCH_V1 and higher, we don't need an event_mappings file
 			if (!fp) {
 				fprintf(stderr, "oprofile: could not open event mapping file %s\n", filename);
 				exit(EXIT_FAILURE);
 			} else {
 				map = get_mapping(nr, fp);
 			}
-			break;			
+			break;
 		default:
 			break;
 	}
@@ -1008,6 +1045,97 @@ struct op_event * op_find_event_any(op_cpu cpu_type, u32 nr)
 	return find_event_any(nr);
 }
 
+static int _is_um_valid_bitmask(struct op_event * event, u32 passed_um)
+{
+	int duped_um[MAX_UNIT_MASK];
+	int retval = 0;
+	u32 masked_val = 0;
+	u32 i, k;
+	int dup_value_used = 0;
+
+	struct op_event evt;
+	struct op_unit_mask * tmp_um = xmalloc(sizeof(struct op_unit_mask));
+	struct op_unit_mask * tmp_um_no_dups = xmalloc(sizeof(struct op_unit_mask));
+	memset(tmp_um, '\0', sizeof(struct op_unit_mask));;
+	memset(tmp_um_no_dups, '\0', sizeof(struct op_unit_mask));
+	memset(duped_um, '\0', sizeof(int) * MAX_UNIT_MASK);
+
+	// First, we make a copy of the event, with just its unit mask values.
+	evt.unit = tmp_um;
+	evt.unit->num = event->unit->num;
+	for (i = 0; i < event->unit->num; i++)
+		evt.unit->um[i].value = event->unit->um[i].value;
+
+	// Next, we sort the unit mask values in ascending order.
+	for (i = 1; i < evt.unit->num; i++) {
+		int j = i - 1;
+		u32 tmp = evt.unit->um[i].value;
+		while (j >= 0 && tmp < evt.unit->um[j].value) {
+			evt.unit->um[j + 1].value = evt.unit->um[j].value;
+			j -= 1;
+		}
+		evt.unit->um[j + 1].value = tmp;
+	}
+
+	/* Now we remove duplicates. Duplicate unit mask values were not
+	 * allowed until the "named unit mask" support was added in
+	 * release 0.9.7.  The down side to this is that if the user passed
+	 * a unit mask value that includes one of the duplicated values,
+	 * we have no way of differentiating between the duplicates, so
+	 * the meaning of the bitmask would be ambiguous if we were to
+	 * allow it.  Thus, we must prevent the user from specifying such
+	 * bitmasks.
+	 */
+	for (i = 0, k = 0; k < evt.unit->num; i++) {
+		tmp_um_no_dups->um[i].value = evt.unit->um[k].value;
+		tmp_um_no_dups->num++;
+		k++;
+		while ((evt.unit->um[i].value == evt.unit->um[k].value) && i < evt.unit->num) {
+			k++;
+			duped_um[i] = 1;
+		}
+	}
+	evt.unit = tmp_um_no_dups;
+
+	// Now check if passed um==0 and if the defined event has a UM with value '0'.
+	if (!passed_um) {
+		for (i = 0; i < evt.unit->num; i++) {
+			if (!evt.unit->um[i].value)
+				return 1;
+		}
+	}
+
+	/* Finally, we'll see if the passed unit mask value can be matched with a
+	 * mask of available unit mask values. We check for this by determining
+	 * whether the exact bits set in the current um are also set in the
+	 * passed um; if so, we OR those bits into a cumulative masked_val variable.
+	 * Simultaneously, we check if the passed um contains a non-unique unit
+	 * mask value, in which case, it's invalid..
+	 */
+	for (i = 0; i < evt.unit->num; i++) {
+		if ((evt.unit->um[i].value & passed_um) == evt.unit->um[i].value) {
+			masked_val |= evt.unit->um[i].value;
+			if (duped_um[i]) {
+				dup_value_used = 1;
+				break;
+			}
+		}
+	}
+
+	if (dup_value_used) {
+		fprintf(stderr, "Ambiguous bitmask: Unit mask values"
+		        " cannot include non-unique numerical values (i.e., 0x%x).\n",
+		        evt.unit->um[i].value);
+		fprintf(stderr, "Use ophelp to see the unit mask values for event %s.\n",
+		        event->name);
+	} else if (masked_val == passed_um && passed_um != 0) {
+		retval = 1;
+	}
+	free(tmp_um);
+	free(tmp_um_no_dups);
+	return retval;
+}
+
 int op_check_events(int ctr, u32 nr, u32 um, op_cpu cpu_type)
 {
 	int ret = OP_INVALID_EVENT;
@@ -1028,21 +1156,15 @@ int op_check_events(int ctr, u32 nr, u32 um, op_cpu cpu_type)
 			ret |= OP_INVALID_COUNTER;
 
 		if (event->unit->unit_type_mask == utm_bitmask) {
-			for (i = 0; i < event->unit->num; ++i)
-				um &= ~(event->unit->um[i].value);			
-			
-			if (um)
+			if (!_is_um_valid_bitmask(event, um))
 				ret |= OP_INVALID_UM;
-			
 		} else {
 			for (i = 0; i < event->unit->num; ++i) {
 				if (event->unit->um[i].value == um)
 					break;
 			}
-			
 			if (i == event->unit->num)
 				ret |= OP_INVALID_UM;
-
 		}
 
 		if (ret == OP_OK_EVENT)
@@ -1078,6 +1200,7 @@ void op_default_event(op_cpu cpu_type, struct op_default_event_descr * descr)
  		case CPU_ATOM:
  		case CPU_CORE_I7:
 		case CPU_NEHALEM:
+		case CPU_HASWELL:
 		case CPU_WESTMERE:
 		case CPU_SANDYBRIDGE:
 		case CPU_IVYBRIDGE:
@@ -1085,6 +1208,7 @@ void op_default_event(op_cpu cpu_type, struct op_default_event_descr * descr)
 		case CPU_FAMILY12H:
 		case CPU_FAMILY14H:
 		case CPU_FAMILY15H:
+		case CPU_AMD64_GENERIC:
 			descr->name = "CPU_CLK_UNHALTED";
 			break;
 
@@ -1142,6 +1266,8 @@ void op_default_event(op_cpu cpu_type, struct op_default_event_descr * descr)
 		case CPU_PPC64_CELL:
 		case CPU_PPC64_POWER7:
 		case CPU_PPC64_IBM_COMPAT_V1:
+		case CPU_PPC64_ARCH_V1:
+		case CPU_PPC64_POWER8:
 			descr->name = "CYCLES";
 			break;
 
@@ -1185,8 +1311,10 @@ void op_default_event(op_cpu cpu_type, struct op_default_event_descr * descr)
 		case CPU_PPC_E300:
 			descr->name = "CPU_CLK";
 			break;
-  	        case CPU_S390_Z10:
-  	        case CPU_S390_Z196:
+
+		case CPU_S390_Z10:
+		case CPU_S390_Z196:
+		case CPU_S390_ZEC12:
  			if (op_get_nr_counters(cpu_type) > 1) {
  				descr->name = "HWSAMPLING";
  				descr->count = 4127518;
@@ -1211,33 +1339,16 @@ void op_default_event(op_cpu cpu_type, struct op_default_event_descr * descr)
 	}
 }
 
-static void extra_check(struct op_event *e, u32 unit_mask)
-{
-	unsigned i;
-	int found = 0;
-
-	for (i = 0; i < e->unit->num; i++)
-		if (e->unit->um[i].value == unit_mask)
-			found++;
-	if (found > 1) {
-		fprintf(stderr,
-"Named unit masks not allowed for events without 'extra:' values.\n"
-"Please specify the numerical value for the unit mask. See 'opcontrol'"
-" man page for more info.\n");
-		exit(EXIT_FAILURE);
-	}
-}
-
-static void another_extra_check(struct op_event *e, char *name, unsigned w)
+static void extra_check(struct op_event *e, char *name, unsigned w)
 {
 	int found;
 	unsigned i;
 
 	if (!e->unit->um[w].extra) {
 		fprintf(stderr,
-"Named unit mask (%s) not allowed for event without 'extra:' values.\n"
-"Please specify the numerical value for the unit mask. See 'opcontrol'"
-" man page for more info.\n", name);
+			"Named unit mask (%s) not allowed for event without 'extra:' values.\n"
+			"Please specify the numerical value for the unit mask. See 'opcontrol'\n"
+			"or 'operf' man page for more info.\n", name);
 		exit(EXIT_FAILURE);
 	}
 
@@ -1248,69 +1359,91 @@ static void another_extra_check(struct op_event *e, char *name, unsigned w)
 		    name[len] == '\0')
 			found++;
 	}
+
 	if (found > 1) {
 		fprintf(stderr,
-	"Unit mask name `%s' not unique. Sorry please use a numerical unit mask\n", name);
+			"Unit mask name `%s' not unique. Please use a numerical unit mask\n", name);
 		exit(EXIT_FAILURE);
 	}
 }
 
-static void do_resolve_unit_mask(struct op_event *e, struct parsed_event *pe,
-				 u32 *extra)
+static void do_resolve_unit_mask(struct op_event *e,
+	struct parsed_event *pe, u32 *extra)
 {
 	unsigned i;
-	int found;
+
+	/* If not specified um and the default um is name type
+	 * we populate pe unitmask name with default name */
+	if ((e->unit->default_mask_name != NULL) &&
+			(pe->unit_mask_name == NULL) && (!pe->unit_mask_valid)) {
+		pe->unit_mask_name = xstrdup(e->unit->default_mask_name);
+	}
 
 	for (;;) {
 		if (pe->unit_mask_name == NULL) {
-			int had_unit_mask = pe->unit_mask_valid;
+			/* For numerical unit mask */
+			int found = 0;
 
-			found = 0;
-			for (i = 0; i < e->unit->num; i++) {
-				if (!pe->unit_mask_valid &&
-				e->unit->um[i].value == e->unit->default_mask) {
-					pe->unit_mask_valid = 1;
-					pe->unit_mask = e->unit->default_mask;
-					break;
-				}
+			/* Use default unitmask if not specified */
+			if (!pe->unit_mask_valid) {
+				pe->unit_mask_valid = 1;
+				pe->unit_mask = e->unit->default_mask;
 			}
-			if (found > 1 && had_unit_mask) {
-				fprintf(stderr,
-	"Non unique numerical unit mask.\n"
-	"Please specify the unit mask using the first word of the description\n");
+
+			/* Checking to see there are any duplicate numerical unit mask
+			 * in which case it should be using named unit mask instead.
+			 */
+			for (i = 0; i < e->unit->num; i++) {
+				if (e->unit->um[i].value == (unsigned int)pe->unit_mask)
+					found++;
+			}
+			if (found > 1) {
+				fprintf(stderr, "Unit mask (0x%x) is non unique.\n"
+				        "Please specify the unit mask using the first "
+					"word of the description\n",
+					pe->unit_mask);
 				exit(EXIT_FAILURE);
 			}
-			extra_check(e, pe->unit_mask);
+
 			if (i == e->unit->num) {
 				e = find_next_event(e);
 				if (e != NULL)
 					continue;
-			} else {
-				if (extra)
+			}
+			return;
+		} else {
+			/* For named unit mask */
+			for (i = 0; i < e->unit->num; i++) {
+				int len = 0;
+
+				if (e->unit->um[i].name)
+					len = strlen(e->unit->um[i].name);
+
+				if (len
+				&&  (!strncmp(pe->unit_mask_name,
+					      e->unit->um[i].name, len))
+				&&  (pe->unit_mask_name[len] == '\0'))
+					break;
+			}
+			if (i == e->unit->num) {
+				e = find_next_event(e);
+				if (e != NULL)
+					continue;
+				fprintf(stderr, "Cannot find unit mask %s for %s\n",
+					pe->unit_mask_name, pe->name);
+				exit(EXIT_FAILURE);
+			}
+			extra_check(e, pe->unit_mask_name, i);
+			pe->unit_mask_valid = 1;
+			pe->unit_mask = e->unit->um[i].value;
+			if (extra) {
+				if (e->unit->um[i].extra == EXTRA_NONE)
+					*extra = e->unit->um[i].value;
+				else
 					*extra = e->unit->um[i].extra;
 			}
 			return;
 		}
-		for (i = 0; i < e->unit->num; i++) {
-			int len = strcspn(e->unit->um[i].desc, " \t");
-			if (!strncmp(pe->unit_mask_name, e->unit->um[i].desc,
-				    len) && pe->unit_mask_name[len] == '\0')
-				break;
-		}
-		if (i == e->unit->num) {
-			e = find_next_event(e);
-			if (e != NULL)
-				continue;
-			fprintf(stderr, "Cannot find unit mask %s for %s\n",
-				pe->unit_mask_name, pe->name);
-			exit(EXIT_FAILURE);
-		}
-		another_extra_check(e, pe->unit_mask_name, i);
-		pe->unit_mask_valid = 1;
-		pe->unit_mask = e->unit->um[i].value;
-		if (extra)
-			*extra = e->unit->um[i].extra;
-		return;
 	}
 }
 
