@@ -10,7 +10,7 @@
  * (C) Copyright IBM Corp. 2011
  *
  * Modified by Maynard Johnson <maynardj@us.ibm.com>
- * (C) Copyright IBM Corporation 2012
+ * (C) Copyright IBM Corporation 2012, 2013
  *
  */
 
@@ -22,6 +22,7 @@
 #include <fcntl.h>
 #include <cverb.h>
 #include <iostream>
+#include <sstream>
 #include "operf_counter.h"
 #include "operf_utils.h"
 #ifdef HAVE_LIBPFM
@@ -47,6 +48,7 @@ extern char * app_name;
 extern pid_t app_PID;
 extern verbose vrecord;
 extern verbose vconvert;
+extern void __set_event_throttled(int index);
 
 using namespace std;
 
@@ -63,26 +65,14 @@ static list<event_t *> unresolved_events;
 static struct operf_transient trans;
 static bool sfile_init_done;
 
-/* The handling of mmap's for a process was a bit tricky to get right, in particular,
- * the handling of what I refer to as "deferred mmap's" -- i.e., when we receive an
- * mmap event for which we've not yet received a comm event (so we don't know app name
- * for the process).  I have left in some debugging code here (compiled out via #ifdef)
- * so we can easily test and validate any changes we ever may need to make to this code.
- */
-//#define _TEST_DEFERRED_MAPPING
-#ifdef _TEST_DEFERRED_MAPPING
-static bool do_comm_event;
-static event_t comm_event;
-#endif
-
-
 /* Some architectures (e.g., ppc64) do not use the same event value (code) for oprofile
  * and for perf_events.  The operf-record process requires event values that perf_events
  * understands, but the operf-read process requires oprofile event values.  The purpose of
  * the following method is to map the operf-record event value to a value that
  * opreport can understand.
  */
-#if (defined(__powerpc__) || defined(__powerpc64__))
+#if PPC64_ARCH
+extern op_cpu cpu_type;
 #define NIL_CODE ~0U
 
 #if HAVE_LIBPFM3
@@ -92,7 +82,6 @@ static bool _get_codes_for_match(unsigned int pfm_idx, const char name[],
 	unsigned int num_events = evt_vec->size();
 	int tmp_code, ret;
 	char evt_name[OP_MAX_EVT_NAME_LEN];
-	char * grp_name;
 	unsigned int events_converted = 0;
 	for (unsigned int i = 0; i < num_events; i++) {
 		operf_event_t event = (*evt_vec)[i];
@@ -103,8 +92,9 @@ static bool _get_codes_for_match(unsigned int pfm_idx, const char name[],
 		memset(evt_name, 0, OP_MAX_EVT_NAME_LEN);
 		if (!strcmp(event.name, "CYCLES")) {
 			strcpy(evt_name ,"PM_CYC") ;
-		} else if ((grp_name = strstr(event.name, "_GRP"))) {
-			strncpy(evt_name, event.name, grp_name - event.name);
+		} else if (strstr(event.name, "_GRP")) {
+			string str = event.name;
+			strncpy(evt_name, event.name, str.rfind("_GRP"));
 		} else {
 			strncpy(evt_name, event.name, strlen(event.name));
 		}
@@ -131,7 +121,6 @@ static bool _op_get_event_codes(vector<operf_event_t> * evt_vec)
 	int ret, i;
 	unsigned int num_events = evt_vec->size();
 	char evt_name[OP_MAX_EVT_NAME_LEN];
-	char * grp_name;
 	unsigned int events_converted = 0;
 	uint64_t code[1];
 
@@ -153,11 +142,16 @@ static bool _op_get_event_codes(vector<operf_event_t> * evt_vec)
 
 	for (unsigned int i = 0; i < num_events; i++) {
 		operf_event_t event = (*evt_vec)[i];
+		if (event.evt_code != NIL_CODE) {
+			events_converted++;
+			continue;
+		}
 		memset(evt_name, 0, OP_MAX_EVT_NAME_LEN);
 		if (!strcmp(event.name, "CYCLES")) {
 			strcpy(evt_name ,"PM_CYC") ;
-		} else if ((grp_name = strstr(event.name, "_GRP"))) {
-			strncpy(evt_name, event.name, grp_name - event.name);
+		} else if (strstr(event.name, "_GRP")) {
+			string str = event.name;
+			strncpy(evt_name, event.name, str.rfind("_GRP"));
 		} else {
 			strncpy(evt_name, event.name, strlen(event.name));
 		}
@@ -188,7 +182,17 @@ bool OP_perf_utils::op_convert_event_vals(vector<operf_event_t> * evt_vec)
 	int ret;
 	for (unsigned int i = 0; i < evt_vec->size(); i++) {
 		operf_event_t event = (*evt_vec)[i];
-		event.evt_code = NIL_CODE;
+		if (cpu_type == CPU_PPC64_POWER7) {
+			if (!strncmp(event.name, "PM_RUN_CYC", strlen("PM_RUN_CYC"))) {
+				event.evt_code = 0x600f4;
+			} else if (!strncmp(event.name, "PM_RUN_INST_CMPL", strlen("PM_RUN_INST_CMPL"))) {
+				event.evt_code = 0x500fa;
+			} else {
+				event.evt_code = NIL_CODE;
+			}
+		} else {
+			event.evt_code = NIL_CODE;
+		}
 		(*evt_vec)[i] = event;
 	}
 
@@ -213,7 +217,7 @@ bool OP_perf_utils::op_convert_event_vals(vector<operf_event_t> * evt_vec)
 #endif
 }
 
-#endif
+#endif // PPC64_ARCH
 
 
 static inline void update_trans_last(struct operf_transient * trans)
@@ -232,7 +236,8 @@ static void __handle_fork_event(event_t * event)
 {
 	if (cverb << vconvert)
 		cout << "PERF_RECORD_FORK for tgid/tid = " << event->fork.pid
-		     << "/" << event->fork.tid << endl;
+		     << "/" << event->fork.tid << "; parent " << event->fork.ppid
+		     << "/" << event->fork.ptid << endl;
 
 	map<pid_t, operf_process_info *>::iterator it;
 	operf_process_info * parent = NULL;
@@ -245,62 +250,68 @@ static void __handle_fork_event(event_t * event)
 		// Create a new proc info object for the parent, but mark it invalid since we have
 		// not yet received a COMM event for this PID.
 		parent = new operf_process_info(event->fork.ppid, app_name ? app_name : NULL,
-		                                                           app_name != NULL, false);
+		                                app_name != NULL, false);
 		if (cverb << vconvert)
-			cout << "Adding new proc info to collection for PID " << event->fork.ppid << endl;
+			cout << "Adding new proc info to collection for parent PID "
+			     << event->fork.ppid << endl;
 		process_map[event->fork.ppid] = parent;
 	}
 
+	 /* If the forked process's pid is the same as the parent's, we simply ignore the FORK
+	 * event. This is because operf_process_info objects are stored in the map collection
+	 * by pid, meaning that the forked process and its parent reference the same
+	 * operf_process_info object.
+	 */
+	if (event->fork.pid == event->fork.ppid)
+		return;
+
 	it = process_map.find(event->fork.pid);
 	if (it == process_map.end()) {
-		forked_proc = new operf_process_info(event->fork.pid,
-		                                     parent->get_app_name().c_str(),
-		                                     parent->is_appname_valid(), parent->is_valid());
+		forked_proc = new operf_process_info(event->fork.pid, NULL, false, false);
 		if (cverb << vconvert)
-			cout << "Adding new proc info to collection for PID " << event->fork.pid << endl;
+			cout << "Adding new proc info to collection for forked PID "
+			     << event->fork.pid << endl;
 		process_map[event->fork.pid] = forked_proc;
-		forked_proc->connect_forked_process_to_parent(parent);
-		parent->add_forked_pid_association(forked_proc);
-		if (cverb << vconvert)
-			cout << "Connecting forked proc " << event->fork.pid << " to parent" << endl;
+		forked_proc->set_fork_info(parent);
 	} else {
-		/* There are two ways that we may get to this point. One way is if
-		 * we've received a COMM event for the forked process before the FORK event.
+		 /*
 		 * Normally, if parent process A forks child process B which then does an exec, we
-		 * first see a FORK event, followed by a COMM event. But apparently there's no
-		 * guarantee in what order these events may be seen by userspace. No matter -- since
-		 * the exec'ed process is now a standalone process (which will get MMAP events
-		 * for all of its mmappings, there's no need to re-associate it back to the parent
-		 * as we do for a non-exec'ed forked process.  So we'll just ignore it.
+		 * first see a FORK event, followed by a COMM event. In this case, the
+		 * operf_process_info created for the forked process is marked as valid.  But there's
+		 * no guarantee what order these events may be seen by userspace -- we could easily
+		 * get MMAP, FORK, and finally a COMM event, which is opposite of "expected". So we
+		 * must handle this.
 		 *
-		 * But the second way that there may be an existing operf_process_info object is if
-		 * a new mmap event (a real MMAP event or a synthesized event (e.g. for hypervisor
-		 * mmapping) occurred for the forked process before a COMM event was received for it.
-		 * In this case, the forked process will be marked invalid until the COMM event
-		 * is received. But if this process does *not* do an exec, there will never be a
-		 * COMM event for it.  Such forked processes should be tightly connected to their
-		 * parent, so we'll go ahead and associate the forked process with its parent.
-		 * If a COMM event comes later for the forked process, we'll disassociate them.
+		 * For a valid operf_process_info, if the forked process pid is unique from that of
+		 * the parent, it implies a COMM event was already received for this forked process.
+		 * Such processes are treated as standalone processes, so we ignore the FORK event.
+		 * For all other cases, if the forked process has not already been associated with
+		 * its parent (i.e., !is_forked()), we go ahead and set that association.
 		 */
+
 		forked_proc = it->second;
-		if (!forked_proc->is_valid()) {
-			forked_proc->connect_forked_process_to_parent(parent);
-			parent->add_forked_pid_association(forked_proc);
+		if (forked_proc->is_valid()) {
+			// Ignore the FORK event
 			if (cverb << vconvert)
-				cout << "Connecting existing incomplete forked proc " << event->fork.pid
-				     << " to parent" << endl;
+				cout << "Forked proc " << event->fork.pid
+				     << " is currently valid (i.e., PERF_RECORD_COMM already received),"
+				     << " so is independent from parent "
+				     << event->fork.ppid << endl;
+			return;
+		}
+
+		if (!forked_proc->is_forked()) {
+			forked_proc->set_fork_info(parent);
+			if (cverb << vconvert)
+				cout << "Set fork info for PID " << event->fork.pid
+				     << " with parent " << event->fork.ppid << endl;
 		}
 	}
 }
 
+
 static void __handle_comm_event(event_t * event)
 {
-#ifdef _TEST_DEFERRED_MAPPING
-	if (!do_comm_event) {
-		comm_event = event;
-		return;
-	}
-#endif
 	if (cverb << vconvert)
 		cout << "PERF_RECORD_COMM for " << event->comm.comm << ", tgid/tid = "
 		     << event->comm.pid << "/" << event->comm.tid << endl;
@@ -308,14 +319,6 @@ static void __handle_comm_event(event_t * event)
 	map<pid_t, operf_process_info *>::iterator it;
 	it = process_map.find(event->comm.pid);
 	if (it == process_map.end()) {
-		/* TODO: Handle system housekeeping tasks.  For certain kinds of processes,
-		 * we will get a COMM event, but never get an MMAP event (e.g, kpsmoused).
-		 * Without receiving an MMAP event, we have no clue whether the name given
-		 * with the COMM event is a full "appname" or not, so the operf_process_info
-		 * is marked invalid.  We end up dropping all samples for such tasks when
-		 * doing a system-wide profile.
-		 */
-
 		/* A COMM event can occur as the result of the app doing a fork/exec,
 		 * where the COMM event is for the forked process.  In that case, we
 		 * pass the event->comm field as the appname argument to the ctor.
@@ -329,34 +332,58 @@ static void __handle_comm_event(event_t * event)
 			appname_arg = event->comm.comm;
 			is_complete_appname = false;
 		}
-		operf_process_info * proc = new operf_process_info(event->comm.pid,appname_arg,
-		                                                   is_complete_appname, true);
+		/* If tid != pid, this may be a forked process for which we've not yet received
+		 * the PERF_RECORD_FORK event, nor have we received any other events for the
+		 * process (e.g., COMM event for parent).  We mark such proc infos as "invalid" so we
+		 * don't falsely attribute samples to a child thread which should, instead,
+		 * be attributed to its parent.  If this is indeed a forked process, we should
+		 * eventually receive a COMM event for the parent (where tid==pid), at which time,
+		 * we'll mark the proc info valid.  If we never receive a COMM event for a parent,
+		 * the proc info will get marked valid during reprocessing so we can attribute
+		 * deferred samples at that time.
+		 */
+
+		bool valid_bit = (event->comm.pid == event->comm.tid);
+		operf_process_info * proc = new operf_process_info(event->comm.pid, appname_arg,
+		                                                   is_complete_appname, valid_bit);
 		if (cverb << vconvert)
 			cout << "Adding new proc info to collection for PID " << event->comm.pid << endl;
 		process_map[event->comm.pid] = proc;
 	} else {
-		if (it->second->is_valid()) {
-			if (it->second->is_forked()) {
-				/* If the operf_process_info object we found was created as a result of
-				 * a FORK event, then it was associated with the parent process and contains
-				 * the parent's appname.  But now we're getting a COMM event for this forked
-				 * process, which means it did an exec, so we need to change the appname
-				 * to the executable associated with this COMM event, which is done via
-				 * calling disassociate_from_parent().
-				 */
-				if (cverb << vconvert)
-					cout << "Disassociating forked proc " << event->comm.pid
-					     << " from parent" << endl;
-				it->second->disassociate_from_parent(event->comm.comm);
-			} else {
-				if (cverb << vconvert)
-					cout << "Received extraneous COMM event for " << event->comm.comm
-					<< ", PID " << event->comm.pid << endl;
+		/* If we reach this point, it means a proc info object for this pid already exists;
+		 * however, if it was created by something other than a "valid" COMM event (e.g., MMAP event),
+		 * its 'valid' bit will be set to false.  NOTE: A "valid" COMM event is one in which
+		 * tid==pid.
+		 *
+		 * We must handle the following situations:
+		 *  o If valid:
+		 *  	- Existing proc info created for a parent (i.e., tid == pid), and the current
+		 *  	  COMM event is for a child -- and we ignore all child COMM events.
+		 *  	- Existing proc info may have invalid appname, so we call set_appname()
+		 *        and see if this COMM event has an appropriate appname.
+		 *
+		 *  o If not valid:
+		 *  	- Existing proc info was created for the parent by an MMAP type of event, and the
+		 *  	  current COMM event is for the parent.
+		 *  	- Existing proc info was created by FORK; now that we have a COMM event for it,
+		 *  	  the process should be treated as a standalone process, so we call
+		 *  	  try_disassociate_from_parent().
+		 */
+		if (!it->second->is_valid()) {
+			// Ignore child COMM events (i.e., pid != tid).
+			if (event->comm.pid == event->comm.tid) {
+				if (it->second->is_forked()) {
+					it->second->try_disassociate_from_parent(event->comm.comm);
+				} else {
+					// Existing proc info created by MMAP event or some such
+					it->second->set_valid();
+					it->second->set_appname(event->comm.comm, false);
+				}
 			}
 		} else {
-			if (cverb << vconvert)
-				cout << "Processing deferred mappings" << endl;
-			it->second->process_deferred_mappings(event->comm.comm);
+			if ((event->comm.pid == event->comm.tid) && !it->second->is_appname_valid()) {
+				it->second->set_appname(event->comm.comm, false);
+			}
 		}
 	}
 }
@@ -398,9 +425,10 @@ static void __handle_mmap_event(event_t * event)
 		mapping->pgoff = event->mmap.pgoff;
 
 		if (cverb << vconvert) {
-			cout << "PERF_RECORD_MMAP for " << event->mmap.filename << endl;
-			cout << "\tstart_addr: " << hex << mapping->start_addr;
-			cout << "; end addr: " << mapping->end_addr << endl;
+			cout << "PERF_RECORD_MMAP for process " << hex << event->mmap.pid << "/"
+			     << event->mmap.tid << ": " << event->mmap.filename << endl;
+			cout << "\tstart_addr: " << hex << mapping->start_addr
+			     << "; end addr: " << mapping->end_addr << endl;
 		}
 
 		if (event->header.misc & PERF_RECORD_MISC_USER)
@@ -457,28 +485,14 @@ static void __handle_mmap_event(event_t * event)
 
 			operf_process_info * proc = new operf_process_info(event->mmap.pid, appname_arg,
 			                                                   is_complete_appname, false);
-			proc->add_deferred_mapping(mapping);
-			if (cverb << vconvert)
-				cout << "Added deferred mapping " << event->mmap.filename
-				      << " for new process_info object" << endl;
 			process_map[event->mmap.pid] = proc;
-#ifdef _TEST_DEFERRED_MAPPING
-			if (!do_comm_event) {
-				do_comm_event = true;
-				__handle_comm_event(comm_event, out);
-			}
-#endif
-		} else if (!it->second->is_valid()) {
-			it->second->add_deferred_mapping(mapping);
-			if (cverb << vconvert)
-				cout << "Added deferred mapping " << event->mmap.filename
-				      << " for existing but incomplete process_info object" << endl;
+			proc->process_mapping(mapping, false);
 		} else {
-			if (cverb << vconvert)
-				cout << "Process mapping for " << event->mmap.filename << " on behalf of "
-				     << event->mmap.pid << endl;
-			it->second->process_new_mapping(mapping);
+			it->second->process_mapping(mapping, false);
 		}
+		if (cverb << vconvert)
+			cout << "Process mapping for " << event->mmap.filename << " on behalf of "
+			<< event->mmap.pid << endl;
 	}
 }
 
@@ -497,22 +511,12 @@ static struct operf_transient * __get_operf_trans(struct sample_data * data, boo
 	} else {
 		// Find operf_process info for data.tgid.
 		std::map<pid_t, operf_process_info *>::const_iterator it = process_map.find(data->pid);
-		if (it != process_map.end() && (it->second->is_appname_valid())) {
+		if (it != process_map.end() && it->second->is_appname_valid()) {
 			proc = it->second;
 		} else {
-			/* This can happen for the following reasons:
-			 *   - We get a sample before getting a COMM or MMAP
-			 *     event for the process being profiled
-			 *   - The COMM event has been processed, but since that
-			 *     only gives 16 chars of the app name, we don't
-			 *     have a valid app name yet
-			 *   - The kernel incorrectly records a sample for a
-			 *     process other than the one we requested (not
-			 *     likely -- this would be a kernel bug if it did)
-			 *
-			*/
+			// This can validly happen if get a sample before getting a COMM event for the process
 			if ((cverb << vconvert) && !first_time_processing) {
-				cerr << "Dropping sample -- process info unavailable" << endl;
+				cout << "Dropping sample -- process info unavailable for PID " << data->pid << endl;
 				if (kernel_mode)
 					operf_stats[OPERF_NO_APP_KERNEL_SAMPLE]++;
 				else
@@ -563,9 +567,10 @@ static struct operf_transient * __get_operf_trans(struct sample_data * data, boo
 			cout << "Found mmap for sample; image_name is " << op_mmap->filename <<
 			" and app name is " << proc->get_app_name() << endl;
 		trans.image_name = op_mmap->filename;
-		trans.app_filename = proc->get_app_name().c_str();
+		trans.app_len = proc->get_app_name().size();
+		strncpy(trans.app_filename, proc->get_app_name().c_str(), trans.app_len);
+		trans.app_filename[trans.app_len] = '\0';
 		trans.image_len = strlen(trans.image_name);
-		trans.app_len = strlen(trans.app_filename);
 		trans.start_addr = op_mmap->start_addr;
 		trans.end_addr = op_mmap->end_addr;
 		trans.tgid = data->pid;
@@ -584,9 +589,11 @@ static struct operf_transient * __get_operf_trans(struct sample_data * data, boo
 	} else {
 		if ((cverb << vconvert) && !first_time_processing) {
 			string domain = trans.in_kernel ? "kernel" : "userspace";
-			cerr << "Discarding " << domain << " sample for process " << data->pid
-			     << " where no appropriate mapping was found. (pc=0x"
-			     << hex << data->ip <<")" << endl;
+			ostringstream message;
+			message << "Discarding " << domain << " sample for process " << data->pid
+			        << " where no appropriate mapping was found. (pc=0x"
+			        << hex << data->ip <<")" << endl;
+			cout << message.str();
 			operf_stats[OPERF_LOST_NO_MAPPING]++;
 		}
 		retval = NULL;
@@ -671,18 +678,38 @@ static void __map_hypervisor_sample(u64 ip, u32 pid)
 	proc->process_hypervisor_mapping(ip);
 }
 
-static void __handle_sample_event(event_t * event, u64 sample_type)
+static int __handle_throttle_event(event_t * event, u64 sample_type)
+{
+	int rc = 0;
+	trans.event = operfRead.get_eventnum_by_perf_event_id(event->throttle.id);
+	if (trans.event >= 0)
+		__set_event_throttled(trans.event);
+	else
+		rc = -1;
+	return rc;
+}
+
+static int __handle_sample_event(event_t * event, u64 sample_type)
 {
 	struct sample_data data;
 	bool found_trans = false;
 	bool in_kernel;
+	int rc = 0;
 	const struct operf_mmap * op_mmap = NULL;
 	bool hypervisor = (event->header.misc == PERF_RECORD_MISC_HYPERVISOR);
 	u64 *array = event->sample.array;
 
+	/* As we extract the various pieces of information from the sample data array,
+	 * if we find that the sample type does not match up with an expected mandatory
+	 * perf_event_sample_format, we consider this as corruption of the sample data
+	 * stream.  Since it wouldn't make sense to continue with suspect data, we quit.
+	 */
 	if (sample_type & PERF_SAMPLE_IP) {
 		data.ip = event->ip.ip;
 		array++;
+	} else {
+		rc = -1;
+		goto done;
 	}
 
 	if (sample_type & PERF_SAMPLE_TID) {
@@ -690,14 +717,21 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 		data.pid = p[0];
 		data.tid = p[1];
 		array++;
+	} else {
+		rc = -1;
+		goto done;
 	}
 
 	data.id = ~0ULL;
 	if (sample_type & PERF_SAMPLE_ID) {
 		data.id = *array;
 		array++;
+	} else {
+		rc = -1;
+		goto done;
 	}
 
+	// PERF_SAMPLE_CPU is optional (see --separate-cpu).
 	if (sample_type & PERF_SAMPLE_CPU) {
 		u_int32_t *p = (u_int32_t *)array;
 		data.cpu = *p;
@@ -708,7 +742,7 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 	} else if (event->header.misc == PERF_RECORD_MISC_USER) {
 		in_kernel = false;
 	}
-#if (defined(__powerpc__) || defined(__powerpc64__))
+#if PPC64_ARCH
 	else if (event->header.misc == PERF_RECORD_MISC_HYPERVISOR) {
 #define MAX_HYPERVISOR_ADDRESS 0xfffffffULL
 		if (data.ip > MAX_HYPERVISOR_ADDRESS) {
@@ -732,18 +766,22 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 			case PERF_RECORD_MISC_HYPERVISOR:
 				domain = "hypervisor";
 				break;
+#if HAVE_PERF_GUEST_MACROS
 			case PERF_RECORD_MISC_GUEST_KERNEL:
 				domain = "guest OS";
 				break;
 			case PERF_RECORD_MISC_GUEST_USER:
 				domain = "guest user";
 				break;
+#endif
 			default:
 				domain = "unknown";
 				break;
 			}
-			cerr << "Discarding sample from " << domain << " domain: "
-			     << hex << data.ip << endl;
+			ostringstream message;
+			message << "Discarding sample from " << domain << " domain: "
+			        << hex << data.ip << endl;
+			cout << message.str();
 		}
 		goto out;
 	}
@@ -759,17 +797,23 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 		goto out;
 	}
 
-	if (cverb << vconvert)
-		cout << "(IP, " <<  event->header.misc << "): " << dec << data.pid << "/"
-		      << data.tid << ": " << hex << (unsigned long long)data.ip
-		      << endl << "\tdata ID: " << data.id << endl;
+	if (cverb << vconvert) {
+		ostringstream message;
+		message << "(IP, " <<  event->header.misc << "): " << dec << data.pid << "/"
+		        << data.tid << ": " << hex << (unsigned long long)data.ip
+		        << endl << "\tdata ID: " << data.id << endl;
+		cout << message.str();
+	}
 
 	// Verify the sample.
-	trans.event = operfRead.get_eventnum_by_perf_event_id(data.id);
-	if (trans.event < 0) {
-		cerr << "Event num " << trans.event << " for id " << data.id
-		     << " is invalid. Skipping sample." << endl;
-		goto out;
+	if (data.id != trans.sample_id) {
+		trans.event = operfRead.get_eventnum_by_perf_event_id(data.id);
+		if (trans.event < 0) {
+			cerr << "Event num " << trans.event << " for id " << data.id
+					<< " is invalid. Sample data appears to be corrupted." << endl;
+			rc = -1;
+			goto out;
+		}
 	}
 
 	/* Only need to check for "no_user" since "no_kernel" is done by
@@ -845,7 +889,7 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 	}
 
 	if (first_time_processing) {
-		event_t * ev = (event_t *)xmalloc(event->header.size);
+		event_t * ev = (event_t *)malloc(event->header.size);
 		memcpy(ev, event, event->header.size);
 		unresolved_events.push_back(ev);
 	}
@@ -853,7 +897,7 @@ static void __handle_sample_event(event_t * event, u64 sample_type)
 out:
 	clear_trans(&trans);
 done:
-	return;
+	return rc;
 }
 
 
@@ -865,8 +909,11 @@ done:
  * when profiling began.  Additional PERF_RECORD_MMAP records may appear later in the data
  * stream (e.g., dlopen for single-process profiling or new process startup for system-wide
  * profiling.
+ *
+ * This function returns '0' on success and '-1' on failure.  A failure implies the sample
+ * data is probably corrupt and the calling function should handle appropriately.
  */
-void OP_perf_utils::op_write_event(event_t * event, u64 sample_type)
+int OP_perf_utils::op_write_event(event_t * event, u64 sample_type)
 {
 #if 0
 	if (event->header.type < PERF_RECORD_MAX) {
@@ -876,47 +923,80 @@ void OP_perf_utils::op_write_event(event_t * event, u64 sample_type)
 
 	switch (event->header.type) {
 	case PERF_RECORD_SAMPLE:
-		__handle_sample_event(event, sample_type);
-		return;
+		return __handle_sample_event(event, sample_type);
 	case PERF_RECORD_MMAP:
 		__handle_mmap_event(event);
-		return;
+		return 0;
 	case PERF_RECORD_COMM:
 		if (!sfile_init_done) {
 			operf_sfile_init();
 			sfile_init_done = true;
 		}
 		__handle_comm_event(event);
-		return;
+		return 0;
 	case PERF_RECORD_FORK:
 		__handle_fork_event(event);
-		return;
+		return 0;
 	case PERF_RECORD_THROTTLE:
-		throttled = true;
-		return;
+		return __handle_throttle_event(event, sample_type);
 	case PERF_RECORD_LOST:
 		operf_stats[OPERF_RECORD_LOST_SAMPLE] += event->lost.lost;
-		return;
+		return 0;
 	case PERF_RECORD_EXIT:
-		return;
+		return 0;
 	default:
-		// OK, ignore all other header types.
-		cverb << vconvert << "No matching event type for " << hex << event->header.type << endl;
-		return;
+		if (event->header.type > PERF_RECORD_MAX) {
+			// Bad header
+			ostringstream message;
+			message << "Invalid event type " << hex << event->header.type << endl;
+			message << "Sample data is probably corrupted." << endl;
+			cerr << message.str();
+			return -1;
+		} else {
+			ostringstream message;
+			message << "Event type "<< hex << event->header.type
+			        << " is ignored." << endl;
+			cverb << vconvert << message.str();
+			return 0;
+		}
 	}
 }
 
-void OP_perf_utils::op_reprocess_unresolved_events(u64 sample_type)
+void OP_perf_utils::op_reprocess_unresolved_events(u64 sample_type, bool print_progress)
 {
+	int num_recs = 0;
+
 	cverb << vconvert << "Reprocessing samples" << endl;
+
+	map<pid_t, operf_process_info *>::iterator procs = process_map.begin();
+	for (; procs != process_map.end(); procs++) {
+		if (!procs->second->is_valid()) {
+			if (procs->second->is_forked()) {
+				procs->second->connect_forked_process_to_parent();
+			} else {
+				procs->second->set_valid();
+			}
+		}
+		// Force the appname_valid to true so we don't drop any samples for this process.
+		// The appname may not be accurate, but it's the best we can do now.
+		procs->second->set_appname_valid();
+	}
 	list<event_t *>::const_iterator it = unresolved_events.begin();
+	int data_error = 0;
 	for (; it != unresolved_events.end(); it++) {
 		event_t * evt = (*it);
+		if (data_error < 0) {
+			free(evt);
+			continue;
+		}
 		// This is just a sanity check, since all events in this list
 		// are unresolved sample events.
 		if (evt->header.type == PERF_RECORD_SAMPLE) {
-			__handle_sample_event(evt, sample_type);
+			data_error = __handle_sample_event(evt, sample_type);
 			free(evt);
+			num_recs++;
+			if ((num_recs % 1000000 == 0) && print_progress)
+				cerr << ".";
 		}
 	}
 }
@@ -946,13 +1026,6 @@ void OP_perf_utils::op_perfrecord_sigusr1_handler(int sig __attribute__((unused)
 	quit = true;
 }
 
-void OP_perf_utils::op_perfread_sigusr1_handler(int sig __attribute__((unused)),
-		siginfo_t * siginfo __attribute__((unused)),
-		void *u_context __attribute__((unused)))
-{
-	read_quit = true;
-}
-
 int OP_perf_utils::op_read_from_stream(ifstream & is, char * buf, streamsize sz)
 {
 	int rc = 0;
@@ -975,13 +1048,18 @@ static int __mmap_trace_file(struct mmap_info & info)
 	info.buf = (char *) mmap(NULL, mmap_size, mmap_prot,
 	                         mmap_flags, info.traceFD, info.offset);
 	if (info.buf == MAP_FAILED) {
-		cerr << "Error: mmap failed with errno:\n\t" << strerror(errno) << endl;
+		ostringstream message;
+		message << "Error: mmap failed with errno:\n\t" << strerror(errno) << endl;
+		message << "\tmmap_size: 0x" << hex << mmap_size << "; offset: 0x" << info.offset << endl;
+		cerr << message.str();
 		return -1;
 	}
 	else {
-		cverb << vconvert << hex << "mmap with the following parameters" << endl
-		      << "\tinfo.head: " << info.head << endl
-		      << "\tinfo.offset: " << info.offset << endl;
+		ostringstream message;
+		message << hex << "mmap with the following parameters" << endl
+		        << "\tinfo.head: " << info.head << endl
+		        << "\tinfo.offset: " << info.offset << endl;
+		cverb << vconvert << message.str();
 		return 0;
 	}
 }
@@ -991,8 +1069,6 @@ int OP_perf_utils::op_mmap_trace_file(struct mmap_info & info, bool init)
 {
 	u64 shift;
 	if (init) {
-		if (!pg_sz)
-			pg_sz = sysconf(_SC_PAGESIZE);
 		if (!mmap_size) {
 			if (MMAP_WINDOW_SZ > info.file_data_size) {
 				mmap_size = info.file_data_size;
@@ -1017,7 +1093,10 @@ int OP_perf_utils::op_write_output(int output, void *buf, size_t size)
 		int ret = write(output, buf, size);
 
 		if (ret < 0) {
-			string errmsg = "Internal error:  Failed to write sample data to pipe. errno is ";
+			if (errno == EINTR)
+				continue;
+
+			string errmsg = "Internal error:  Failed to write sample data to output fd. errno is ";
 			errmsg += strerror(errno);
 			throw runtime_error(errmsg);
 		}
@@ -1030,7 +1109,7 @@ int OP_perf_utils::op_write_output(int output, void *buf, size_t size)
 }
 
 
-static void op_record_process_exec_mmaps(pid_t pid, pid_t tgid, int output_fd, operf_record * pr)
+void OP_perf_utils::op_record_process_exec_mmaps(pid_t pid, pid_t tgid, int output_fd, operf_record * pr)
 {
 	char fname[PATH_MAX];
 	FILE *fp;
@@ -1048,6 +1127,8 @@ static void op_record_process_exec_mmaps(pid_t pid, pid_t tgid, int output_fd, o
 		char line_buffer[BUFSIZ];
 		char perms[5], pathname[PATH_MAX], dev[16];
 		unsigned long long start_addr, end_addr, offset;
+		const char * anon_mem = "//anon";
+
 		u_int32_t inode;
 
 		memset(pathname, '\0', sizeof(pathname));
@@ -1069,6 +1150,9 @@ static void op_record_process_exec_mmaps(pid_t pid, pid_t tgid, int output_fd, o
 			if (imagename == NULL)
 				imagename = strstr(pathname, "[vdso]");
 
+			if ((imagename == NULL) && !strstr(pathname, "["))
+				imagename = (char *)anon_mem;
+
 			if (imagename == NULL)
 				continue;
 
@@ -1092,8 +1176,7 @@ static void op_record_process_exec_mmaps(pid_t pid, pid_t tgid, int output_fd, o
 	return;
 }
 
-static int _record_one_process_info(pid_t pid, bool sys_wide, operf_record * pr,
-                                    int output_fd)
+static int _get_one_process_info(bool sys_wide, pid_t pid, operf_record * pr)
 {
 	struct comm_event comm;
 	char fname[PATH_MAX];
@@ -1116,7 +1199,7 @@ static int _record_one_process_info(pid_t pid, bool sys_wide, operf_record * pr,
 		if (!sys_wide) {
 			cerr << "Unable to find process information for process " << pid << "." << endl;
 			cverb << vrecord << "couldn't open " << fname << endl;
-			return -1;
+			return OP_PERF_HANDLED_ERROR;
 		} else {
 			return 0;
 		}
@@ -1151,10 +1234,10 @@ static int _record_one_process_info(pid_t pid, bool sys_wide, operf_record * pr,
 	size = align_64bit(size);
 	comm.header.size = sizeof(comm) - (sizeof(comm.comm) - size);
 	if (tgid != pid) {
-		// passed pid must have been a secondary thread
+		// passed pid must have been a secondary thread, and we
+		// don't go looking at the /proc/<pid>/task of such processes.
 		comm.tid = pid;
-		int num = OP_perf_utils::op_write_output(output_fd, &comm, comm.header.size);
-		pr->add_to_total(num);
+		pr->add_process(comm);
 		goto out;
 	}
 
@@ -1163,7 +1246,8 @@ static int _record_one_process_info(pid_t pid, bool sys_wide, operf_record * pr,
 	if (tids == NULL) {
 		// process must have exited
 		ret = -1;
-		cverb << vrecord << "opendir returned NULL" << endl;
+		cverb << vrecord << "Process " << pid << " apparently exited while "
+		      << "process info was being collected"<< endl;
 		goto out;
 	}
 
@@ -1174,39 +1258,30 @@ static int _record_one_process_info(pid_t pid, bool sys_wide, operf_record * pr,
 			continue;
 
 		comm.tid = pid;
-
-		int num = OP_perf_utils::op_write_output(output_fd, &comm, comm.header.size);
-		pr->add_to_total(num);
+		pr->add_process(comm);
 	}
 	closedir(tids);
-	if (cverb << vrecord)
-		cout << "Created COMM event for " << comm.comm << endl;
 
 out:
-	op_record_process_exec_mmaps(pid, tgid, output_fd, pr);
-
 	fclose(fp);
 	if (ret) {
 		cverb << vrecord << "couldn't get app name and tgid for pid "
 		      << dec << pid << " from /proc fs." << endl;
 	}
 	return ret;
-
 }
 
 /* Obtain process information for an active process (where the user has
  * passed in a process ID via the --pid option) or all active processes
- * (where system_wide==true).  Then generate the necessary PERF_RECORD_COMM
- * and PERF_RECORD_MMAP entries into the profile data stream.
+ * (where system_wide==true).
  */
-int OP_perf_utils::op_record_process_info(bool system_wide, pid_t pid, operf_record * pr,
-                                          int output_fd)
+int OP_perf_utils::op_get_process_info(bool system_wide, pid_t pid, operf_record * pr)
 {
 	int ret = 0;
 	if (cverb << vrecord)
-		cout << "op_record_process_info" << endl;
+		cout << "op_get_process_info" << endl;
 	if (!system_wide) {
-		ret = _record_one_process_info(pid, system_wide, pr, output_fd);
+		ret = _get_one_process_info(system_wide, pid, pr);
 	} else {
 		char buff[BUFSIZ];
 		pid_t tgid = 0;
@@ -1228,7 +1303,7 @@ int OP_perf_utils::op_record_process_info(bool system_wide, pid_t pid, operf_rec
 				cverb << vmisc << "/proc entry " << dirent.d_name << " is not a PID" << endl;
 				continue;
 			}
-			if ((ret = _record_one_process_info(pid, system_wide, pr, output_fd)) < 0)
+			if ((ret = _get_one_process_info(system_wide, pid, pr)) < 0)
 				break;
 		}
 		closedir(pids);
@@ -1337,9 +1412,12 @@ void OP_perf_utils::op_record_kernel_info(string vmlinux_file, u64 start_addr, u
 	mmap.header.size = (sizeof(mmap) -
 			(sizeof(mmap.filename) - size));
 	int num = op_write_output(output_fd, &mmap, mmap.header.size);
-	if (cverb << vrecord)
-		cout << "Created MMAP event of size " << mmap.header.size << " for " <<mmap.filename << ". length: "
-		     << hex << mmap.len << "; start addr: " << mmap.start << endl;
+	if (cverb << vrecord) {
+		ostringstream message;
+		message << "Created MMAP event of size " << mmap.header.size << " for " <<mmap.filename << ". length: "
+		        << hex << mmap.len << "; start addr: " << mmap.start << endl;
+		cout << message.str();
+	}
 	pr->add_to_total(num);
 	_record_module_info(output_fd, pr);
 }
@@ -1360,6 +1438,9 @@ void OP_perf_utils::op_get_kernel_event_data(struct mmap_data *md, operf_record
 	void *buf;
 	int64_t diff;
 
+	if (old == head)
+		return;
+
 	diff = head - old;
 	if (diff < 0) {
 		throw runtime_error("ERROR: event buffer wrapped, which should NEVER happen.");
