@@ -11,7 +11,7 @@
  * (C) Copyright IBM Corp. 2011
  *
  * Modified by Maynard Johnson <maynardj@us.ibm.com>
- * (C) Copyright IBM Corporation 2012
+ * (C) Copyright IBM Corporation 2012, 2013
  *
  */
 
@@ -46,6 +46,8 @@
 #include "operf_kernel.h"
 #include "child_reader.h"
 #include "op_get_time.h"
+#include "operf_stats.h"
+#include "op_netburst.h"
 
 using namespace std;
 
@@ -59,13 +61,13 @@ typedef enum END_CODE {
 
 // Globals
 char * app_name = NULL;
+bool use_cpu_minus_one = false;
 pid_t app_PID = -1;
 uint64_t kernel_start, kernel_end;
 operf_read operfRead;
 op_cpu cpu_type;
 double cpu_speed;
-char op_samples_current_dir[PATH_MAX];
-uint op_nr_counters;
+uint op_nr_events;
 verbose vmisc("misc");
 uid_t my_uid;
 bool no_vmlinux;
@@ -77,7 +79,7 @@ char * start_time_human_readable;
 
 static char full_pathname[PATH_MAX];
 static char * app_name_SAVE = NULL;
-static char * app_args = NULL;
+static char ** app_args = NULL;
 static 	pid_t jitconv_pid = -1;
 static bool app_started;
 static pid_t operf_record_pid;
@@ -90,7 +92,8 @@ static vector<operf_event_t> events;
 static bool jit_conversion_running;
 static void convert_sample_data(void);
 static int sample_data_pipe[2];
-static bool ctl_c = false;
+bool ctl_c = false;
+bool pipe_closed = false;
 
 
 namespace operf_options {
@@ -133,6 +136,16 @@ const char * short_options = "V:d:k:gsap:e:ctlhuv";
 
 vector<string> verbose_string;
 
+void __set_event_throttled(int index)
+{
+	if (index < 0) {
+		cerr << "Unable to determine if throttling occurred for ";
+		cerr << "event " << events[index].name << endl;
+	} else {
+		events[index].throttled = true;
+	}
+}
+
 static void __print_usage_and_exit(const char * extra_msg)
 {
 	if (extra_msg)
@@ -142,7 +155,7 @@ static void __print_usage_and_exit(const char * extra_msg)
 	exit(EXIT_FAILURE);
 }
 
-
+// Signal handler for main (parent) process.
 static void op_sig_stop(int val __attribute__((unused)))
 {
 	// Received a signal to quit, so we need to stop the
@@ -155,9 +168,14 @@ static void op_sig_stop(int val __attribute__((unused)))
 		kill(app_PID, SIGKILL);
 }
 
+// For child processes to manage a controlled stop after Ctl-C is done
 static void _handle_sigint(int val __attribute__((unused)))
 {
 	size_t dummy __attribute__ ((__unused__));
+	/* Each process (parent and each forked child) will have their own copy of
+	 * the ctl_c variable, so this can be used by each process in managing their
+	 * shutdown procedure.
+	 */
 	ctl_c = true;
 	if (cverb << vdebug)
 		dummy = write(1, "in _handle_sigint\n", 19);
@@ -165,7 +183,7 @@ static void _handle_sigint(int val __attribute__((unused)))
 }
 
 
-void _set_signals_for_record(void)
+void _set_basic_SIGINT_handler_for_child(void)
 {
 	struct sigaction act;
 	sigset_t ss;
@@ -183,7 +201,7 @@ void _set_signals_for_record(void)
 	}
 }
 
-void set_signals(void)
+void set_signals_for_parent(void)
 {
 	struct sigaction act;
 	sigset_t ss;
@@ -204,47 +222,28 @@ void set_signals(void)
 
 static int app_ready_pipe[2], start_app_pipe[2], operf_record_ready_pipe[2];
 
+static string args_to_string(void)
+{
+	string ret;
+	char * const * ptr = app_args + 1;
+	while (*ptr != NULL) {
+		ret.append(*ptr);
+		ret += ' ';
+		ptr++;
+	}
+	return ret;
+}
+
 void run_app(void)
 {
+	// ASSUMPTION: app_name is a fully-qualified pathname
 	char * app_fname = rindex(app_name, '/') + 1;
-	if (!app_fname) {
-		string msg = "Error trying to parse app name ";
-		msg += app_name;
-		__print_usage_and_exit(msg.c_str());
-	}
+	app_args[0] = app_fname;
 
-	vector<string> exec_args_str;
-	if (app_args) {
-		size_t end_pos;
-		string app_args_str = app_args;
-		// Since the strings returned from substr would otherwise be ephemeral, we
-		// need to store them into the exec_args_str vector so we can reference
-		// them later when we call execvp.
-		do {
-			end_pos = app_args_str.find_first_of(' ', 0);
-			if (end_pos != string::npos) {
-				exec_args_str.push_back(app_args_str.substr(0, end_pos));
-				app_args_str = app_args_str.substr(end_pos + 1);
-			} else {
-				exec_args_str.push_back(app_args_str);
-			}
-		} while (end_pos != string::npos);
-	}
-
-	vector<const char *> exec_args;
-	exec_args.push_back(app_fname);
-	vector<string>::iterator it;
-	cverb << vdebug << "Exec args are: " << app_fname << " ";
-	// Now transfer the args from the intermediate exec_args_str container to the
-	// exec_args container that can be passed to execvp.
-	for (it = exec_args_str.begin(); it != exec_args_str.end(); it++) {
-		exec_args.push_back((*it).c_str());
-		cverb << vdebug << (*it).c_str() << " ";
-	}
-	exec_args.push_back((char *) NULL);
-	cverb << vdebug << endl;
+	string arg_str = args_to_string();
+	cverb << vdebug << "Exec args are: " << app_fname << " " << arg_str << endl;
 	// Fake an exec to warm-up the resolver
-	execvp("", ((char * const *)&exec_args[0]));
+	execvp("", app_args);
 	// signal to the parent that we're ready to exec
 	int startup = 1;
 	if (write(app_ready_pipe[1], &startup, sizeof(startup)) < 0) {
@@ -263,8 +262,8 @@ void run_app(void)
 
 	cverb << vdebug << "parent says start app " << app_name << endl;
 	app_started = true;
-	execvp(app_name, ((char * const *)&exec_args[0]));
-	cerr <<  "Failed to exec " << exec_args[0] << ": " << strerror(errno) << endl;
+	execvp(app_name, app_args);
+	cerr <<  "Failed to exec " << app_fname << " " << arg_str << ": " << strerror(errno) << endl;
 	/* We don't want any cleanup in the child */
 	_exit(EXIT_FAILURE);
 
@@ -312,7 +311,7 @@ int start_profiling(void)
 	} else if (operf_record_pid == 0) { // operf-record process
 		int ready = 0;
 		int exit_code = EXIT_SUCCESS;
-		_set_signals_for_record();
+		_set_basic_SIGINT_handler_for_child();
 		close(operf_record_ready_pipe[0]);
 		if (!operf_options::post_conversion)
 			close(sample_data_pipe[0]);
@@ -367,7 +366,7 @@ int start_profiling(void)
 
 			// start recording
 			operfRecord.recordPerfData();
-			cverb << vmisc << "Total bytes recorded from perf events: " << dec
+			cverb << vdebug << "Total bytes recorded from perf events: " << dec
 					<< operfRecord.get_total_bytes_recorded() << endl;
 		} catch (runtime_error re) {
 			/* If the user does ctl-c, the operf-record process may get interrupted
@@ -493,7 +492,7 @@ static end_code_t _kill_operf_read_pid(end_code_t rc)
 				if (!WEXITSTATUS(waitpid_status)) {
 					cverb << vdebug << "operf-read process returned OK" << endl;
 				} else if (WIFEXITED(waitpid_status)) {
-					/* If user did ctl-c, operf-record may get spurious errors, like
+					/* If user did ctl-c, operf-read may get spurious errors, like
 					 * broken pipe, etc.  We ignore these unless the user asks for
 					 * debug output.
 					 */
@@ -505,7 +504,7 @@ static end_code_t _kill_operf_read_pid(end_code_t rc)
 				}
 			}  else if (WIFSIGNALED(waitpid_status)) {
 				keep_trying = false;
-				/* If user did ctl-c, operf-record may get spurious errors, like
+				/* If user did ctl-c, operf-read may get spurious errors, like
 				 * broken pipe, etc.  We ignore these unless the user asks for
 				 * debug output.
 				 */
@@ -526,9 +525,13 @@ static end_code_t _kill_operf_record_pid(void)
 	end_code_t rc = ALL_OK;
 
 	// stop operf-record process
+	errno = 0;
 	if (kill(operf_record_pid, SIGUSR1) < 0) {
-		perror("Attempt to stop operf-record process failed");
-		rc = PERF_RECORD_ERROR;
+		// If operf-record process is already ended, don't consider this an error.
+		if (errno != ESRCH) {
+			perror("Attempt to stop operf-record process failed");
+			rc = PERF_RECORD_ERROR;
+		}
 	} else {
 		if (waitpid(operf_record_pid, &waitpid_status, 0) < 0) {
 			perror("waitpid for operf-record process failed");
@@ -587,7 +590,7 @@ static end_code_t _run(void)
 	/* If we're not doing system wide profiling and no app is started, then
 	 * there's no profile data to convert. So if this condition is NOT true,
 	 * then we'll do the convert.
-	 * Note that if --lazy-connversion is passed, then operf_options::post_conversion
+	 * Note that if --lazy-conversion is passed, then operf_options::post_conversion
 	 * will be set, and we will defer conversion until after the operf-record
 	 * process is done.
 	 */
@@ -600,7 +603,9 @@ static end_code_t _run(void)
 				_exit(EXIT_FAILURE);
 			} else if (operf_read_pid == 0) { // child process
 				close(sample_data_pipe[1]);
+				_set_basic_SIGINT_handler_for_child();
 				convert_sample_data();
+				_exit(EXIT_SUCCESS);
 			}
 			// parent
 			close(sample_data_pipe[0]);
@@ -608,8 +613,7 @@ static end_code_t _run(void)
 		}
 	}
 
-	set_signals();
-	cerr << "operf: Profiler started" << endl;
+	set_signals_for_parent();
 	if (startApp) {
 		/* The user passed in a command or program name to start, so we'll need to do waitpid on that
 		 * process.  However, while that user-requested process is running, it's possible we
@@ -619,7 +623,7 @@ static end_code_t _run(void)
 		 * process, checking their status.  The profiled app may end normally, abnormally, or by way
 		 * of ctrl-C.  The operf-record process should not end here, except abnormally.  The normal
 		 * flow is:
-		 *    1. profiled app ends or is stopped vi ctrl-C
+		 *    1. profiled app ends or is stopped via ctrl-C
 		 *    2. keep_trying is set to false, so we drop out of while loop and proceed to end of function
 		 *    3. call _kill_operf_record_pid and _kill_operf_read_pid
 		 */
@@ -779,14 +783,6 @@ static void _set_signals_for_convert(void)
 		perror("operf: install of SIGCHLD handler failed: ");
 		exit(EXIT_FAILURE);
 	}
-
-	act.sa_handler = _handle_sigint;
-	sigemptyset(&act.sa_mask);
-	sigaddset(&act.sa_mask, SIGINT);
-	if (sigaction(SIGINT, &act, NULL)) {
-		perror("operf: install of SIGINT handler failed: ");
-		exit(EXIT_FAILURE);
-	}
 }
 
 static void _do_jitdump_convert()
@@ -850,9 +846,22 @@ static int __delete_old_previous_sample_data(const char *fpath,
 	}
 }
 
-/* Read perf_events sample data written by the operf-record process
- * through the sample_data_pipe and convert this to oprofile format
- * sample files.
+/* Read perf_events sample data written by the operf-record process through
+ * the sample_data_pipe or file (dependent on 'lazy-conversion' option)
+ * and convert the perf format sample data to to oprofile format sample files.
+ *
+ * If not invoked with --lazy-conversion option, this function is executed by
+ * the "operf-read" child process.  If user does a ctrl-C, the parent will
+ * execute _kill_operf_read_pid which will try to allow the conversion process
+ * to complete, waiting 5 seconds before it forcefully kills the operf-read
+ * process via 'kill SIGUSR1'.
+ *
+ * But if --lazy-conversion option is used, then it's the parent process that's
+ * running convert_sample_data.  If the user does a ctrl-C during this procedure,
+ * the ctrl-C is handled via op_sig_stop which essentially does nothing to stop
+ * the conversion procedure, which in general is fine.  On the very rare chance
+ * that the procedure gets stuck (hung) somehow, the user will have to do a
+ * 'kill -KILL'.
  */
 static void convert_sample_data(void)
 {
@@ -862,6 +871,7 @@ static void convert_sample_data(void)
 	int keep_waiting = 0;
 	string current_sampledir = samples_dir + "/current/";
 	string previous_sampledir = samples_dir + "/previous";
+	string stats_dir = "";
 	current_sampledir.copy(op_samples_current_dir, current_sampledir.length(), 0);
 
 	if (!app_started && !operf_options::system_wide)
@@ -912,7 +922,7 @@ static void convert_sample_data(void)
 	cverb << vdebug << "Successfully read header info for sample data " << endl;
 	if (operfRead.is_valid()) {
 		try {
-			int num = operfRead.convertPerfData();
+			unsigned int num = operfRead.convertPerfData();
 			cverb << vdebug << "operf_read: Total bytes received from operf_record process: " << dec << num << endl;
 		} catch (runtime_error e) {
 			cerr << "Caught runtime error from operf_read::convertPerfData" << endl;
@@ -921,6 +931,7 @@ static void convert_sample_data(void)
 			goto out;
 		}
 	}
+
 	_set_signals_for_convert();
 	cverb << vdebug << "Calling _do_jitdump_convert" << endl;
 	_do_jitdump_convert();
@@ -956,8 +967,11 @@ static int get_PATH_based_pathname(char * path_holder, size_t n)
 		struct dirent ** namelist;
 		int rc = scandir(segment, &namelist, find_app_file_in_dir, NULL);
 		if (rc < 0) {
-			cerr << app_name << " cannot be found in your PATH." << endl;
-			break;
+			if (errno != ENOENT) {
+				cerr << strerror(errno) << endl;
+				cerr << app_name << " cannot be found in your PATH." << endl;
+				break;
+			}
 		} else if (rc == 1) {
 			size_t applen = strlen(app_name);
 			size_t dirlen = strlen(segment);
@@ -969,12 +983,23 @@ static int get_PATH_based_pathname(char * path_holder, size_t n)
 				retval = -1;
 				break;
 			}
-			strncpy(path_holder, segment, dirlen);
+
+			if (!strcmp(segment, ".")) {
+				if (getcwd(path_holder, PATH_MAX) == NULL) {
+					retval = -1;
+					cerr << "getcwd [3] failed when processing <cur-dir>/" << app_name << " found via PATH. Aborting."
+							<< endl;
+					break;
+				}
+			} else {
+				strncpy(path_holder, segment, dirlen);
+			}
 			strcat(path_holder, "/");
 			strncat(path_holder, app_name, applen);
 			retval = 0;
 			free(namelist[0]);
 			free(namelist);
+
 			break;
 		}
 		segment = strtok(NULL, ":");
@@ -1008,6 +1033,13 @@ int validate_app_name(void)
 			goto out;
 		}
 		strcat(full_pathname, "/");
+		if ((strlen(full_pathname) + strlen(app_name + 2) + 1) > PATH_MAX) {
+			rc = -1;
+			cerr << "Length of current dir (" << full_pathname << ") and app name ("
+			     << (app_name + 2) << ") exceeds max allowed (" << PATH_MAX << "). Aborting."
+			     << endl;
+			goto out;
+		}
 		strcat(full_pathname, (app_name + 2));
 	} else if (index(app_name, '/')) {
 		// Passed app is in a subdirectory of cur dir; e.g., "test-stuff/myApp"
@@ -1020,7 +1052,7 @@ int validate_app_name(void)
 		strcat(full_pathname, "/");
 		strcat(full_pathname, app_name);
 	} else {
-		// Pass app name, at this point, MUST be found in PATH
+		// Passed app name, at this point, MUST be found in PATH
 		rc = get_PATH_based_pathname(full_pathname, PATH_MAX);
 	}
 
@@ -1048,6 +1080,12 @@ static void _get_event_code(operf_event_t * event)
 	char oprof_event_code[9];
 	string command;
 	u64 base_code, config;
+	char buf[20];
+	if ((snprintf(buf, 20, "%lu", event->count)) < 0) {
+		cerr << "Error parsing event count of " << event->count << endl;
+		exit(EXIT_FAILURE);
+	}
+
 	base_code = config = 0ULL;
 
 	command = OP_BINDIR;
@@ -1074,6 +1112,7 @@ static void _get_event_code(operf_event_t * event)
 
 #if defined(__i386__) || defined(__x86_64__)
 	// Setup EventSelct[11:8] field for AMD
+	char mask[12];
 	const char * vendor_AMD = "AuthenticAMD";
 	if (op_is_cpu_vendor((char *)vendor_AMD)) {
 		config = base_code & 0xF00ULL;
@@ -1084,13 +1123,8 @@ static void _get_event_code(operf_event_t * event)
 	config |= base_code & 0xFFULL;
 
 	// Setup unitmask field
+handle_named_um:
 	if (event->um_name[0]) {
-		char mask[12];
-		char buf[20];
-		if ((snprintf(buf, 20, "%lu", event->count)) == -1) {
-			cerr << "Error parsing event count of " << event->count << endl;
-			exit(EXIT_FAILURE);
-		}
 		command = OP_BINDIR;
 		command += "ophelp ";
 		command += "--extra-mask ";
@@ -1112,7 +1146,48 @@ static void _get_event_code(operf_event_t * event)
 			exit(EXIT_FAILURE);
 		}
 		pclose(fp);
-		config |= strtoull(mask, (char **) NULL, 10);
+		// FIXME:  The mask value here is the extra bits from the named unit mask.  It's not
+		// ideal to put that value into the UM's mask, since that's what will show up in
+		// opreport.  It would be better if we could somehow have the unit mask name that the
+		// user passed to us show up in opreort.
+		event->evt_um = strtoull(mask, (char **) NULL, 10);
+		/* A value >= EXTRA_MIN_VAL returned by 'ophelp --extra-mask' is interpreted as a
+		 * valid extra value; otherwise we interpret it as a simple unit mask value
+		 * for a named unit mask with EXTRA_NONE.
+		 */
+		if (event->evt_um >= EXTRA_MIN_VAL)
+			config |= event->evt_um;
+		else
+			config |= ((event->evt_um & 0xFFULL) << 8);
+	} else if (!event->evt_um) {
+		char * endptr;
+		command.clear();
+		command = OP_BINDIR;
+		command += "ophelp ";
+		command += "--unit-mask ";
+		command += event->name;
+		command += ":";
+		command += buf;
+		fp = popen(command.c_str(), "r");
+		if (fp == NULL) {
+			cerr << "Unable to execute ophelp to get unit mask for event "
+			     << event->name << endl;
+			exit(EXIT_FAILURE);
+		}
+		if (fgets(mask, sizeof(mask), fp) == NULL) {
+			pclose(fp);
+			cerr << "Unable to find unit mask info for event " << event->name << endl;
+			exit(EXIT_FAILURE);
+		}
+		pclose(fp);
+		event->evt_um = strtoull(mask, &endptr, 10);
+		if ((endptr >= mask) &&
+				(endptr <= (mask + strlen(mask) - 1))) {
+			// Must be a default named unit mask
+			strncpy(event->um_name, mask, OP_MAX_UM_NAME_LEN);
+			goto handle_named_um;
+		}
+		config |= ((event->evt_um & 0xFFULL) << 8);
 	} else {
 		config |= ((event->evt_um & 0xFFULL) << 8);
 	}
@@ -1121,17 +1196,137 @@ static void _get_event_code(operf_event_t * event)
 #endif
 
 	event->op_evt_code = base_code;
+	if (cpu_type == CPU_P4 || cpu_type == CPU_P4_HT2) {
+		if (op_netburst_get_perf_encoding(event->name, event->evt_um, 1, 1, &config)) {
+			cerr << "Unable to get event encoding for " << event->name << endl;
+			exit(EXIT_FAILURE);
+		}
+	}
 	event->evt_code = config;
 }
 
+#if PPC64_ARCH
+/* All ppc64 events (except CYCLES) have a _GRP<n> suffix.  This is
+ * because the legacy opcontrol profiler can only profile events in
+ * the same group (i.e., having the same _GRP<n> suffix).  But operf
+ * can multiplex events, so we should allow the user to pass event
+ * names without the _GRP<n> suffix.
+ *
+ * If event name is not CYCLES or does not have a _GRP<n> suffix,
+ * we'll call ophelp and scan the list of events, searching for one
+ * that matches up to the _GRP<n> suffix.  If we don't find a match,
+ * then we'll exit with the expected error message for invalid event name.
+ */
+static string _handle_powerpc_event_spec(string event_spec)
+{
+	FILE * fp;
+	char line[MAX_INPUT];
+	size_t grp_pos;
+	string evt, retval, err_msg;
+	size_t evt_name_len;
+	bool first_non_cyc_evt_found = false;
+	bool event_found = false;
+	char event_name[OP_MAX_EVT_NAME_LEN], event_spec_str[OP_MAX_EVT_NAME_LEN + 20], * count_str;
+	string cmd = OP_BINDIR;
+	cmd += "/ophelp";
+
+	strncpy(event_spec_str, event_spec.c_str(), event_spec.length() + 1);
+
+	strncpy(event_name, strtok(event_spec_str, ":"), OP_MAX_EVT_NAME_LEN);
+	count_str = strtok(NULL, ":");
+	if (!count_str) {
+		err_msg = "Invalid count for event ";
+		goto out;
+	}
+
+	if (!strcmp("CYCLES", event_name)) {
+		event_found = true;
+		goto out;
+	}
+
+	evt = event_name;
+	// Need to make sure the event name truly has a _GRP<n> suffix.
+	grp_pos = evt.rfind("_GRP");
+	if ((grp_pos != string::npos) && ((evt = evt.substr(grp_pos, string::npos))).length() > 4) {
+		char * end;
+		strtoul(evt.substr(4, string::npos).c_str(), &end, 0);
+		if (end && (*end == '\0')) {
+		// Valid group number found after _GRP, so we can skip to the end.
+			event_found = true;
+			goto out;
+		}
+	}
+
+	// If we get here, it implies the user passed a non-CYCLES event without a GRP suffix.
+	// Lets try to find a valid suffix for it.
+	fp = popen(cmd.c_str(), "r");
+	if (fp == NULL) {
+		cerr << "Unable to execute ophelp to get info for event "
+		     << event_spec << endl;
+		exit(EXIT_FAILURE);
+	}
+	evt_name_len = strlen(event_name);
+	err_msg = "Cannot find event ";
+	while (fgets(line, MAX_INPUT, fp)) {
+		if (!first_non_cyc_evt_found) {
+			if (!strncmp(line, "PM_", 3))
+				first_non_cyc_evt_found = true;
+			else
+				continue;
+		}
+		if (line[0] == ' ' || line[0] == '\t')
+			continue;
+		if (!strncmp(line, event_name, evt_name_len)) {
+			// Found a potential match.  Check if it's a perfect match.
+			string save_event_name = event_name;
+			size_t full_evt_len = index(line, ':') - line;
+			memset(event_name, '\0', OP_MAX_EVT_NAME_LEN);
+			strncpy(event_name, line, full_evt_len);
+			string candidate = event_name;
+			if (candidate.rfind("_GRP") == evt_name_len) {
+				event_found = true;
+				break;
+			} else {
+				memset(event_name, '\0', OP_MAX_EVT_NAME_LEN);
+				strncpy(event_name, save_event_name.c_str(), evt_name_len);
+			}
+		}
+	}
+	pclose(fp);
+
+out:
+	if (!event_found) {
+		cerr << err_msg << event_name << endl;
+		cerr << "Error retrieving info for event "
+				<< event_spec << endl;
+		exit(EXIT_FAILURE);
+	}
+	retval = event_name;
+	return retval + ":" + count_str;
+}
+#endif
+
 static void _process_events_list(void)
 {
 	string cmd = OP_BINDIR;
+	if (operf_options::evts.size() > OP_MAX_EVENTS) {
+		cerr << "Number of events specified is greater than allowed maximum of "
+		     << OP_MAX_EVENTS << "." << endl;
+		exit(EXIT_FAILURE);
+	}
 	cmd += "/ophelp --check-events ";
 	for (unsigned int i = 0; i <  operf_options::evts.size(); i++) {
 		FILE * fp;
 		string full_cmd = cmd;
 		string event_spec = operf_options::evts[i];
+
+#if PPC64_ARCH
+		// Starting with CPU_PPC64_ARCH_V1, ppc64 events files are formatted like
+		// other architectures, so no special handling is needed.
+		if (cpu_type < CPU_PPC64_ARCH_V1)
+			event_spec = _handle_powerpc_event_spec(event_spec);
+#endif
+
 		if (operf_options::callgraph) {
 			full_cmd += " --callgraph=1 ";
 		}
@@ -1151,10 +1346,10 @@ static void _process_events_list(void)
 				     << endl << "15 times the minimum count value for the event."  << endl;
 			exit(EXIT_FAILURE);
 		}
-		fclose(fp);
+		pclose(fp);
 		char * event_str = op_xstrndup(event_spec.c_str(), event_spec.length());
 		operf_event_t event;
-		strncpy(event.name, strtok(event_str, ":"), OP_MAX_EVT_NAME_LEN);
+		strncpy(event.name, strtok(event_str, ":"), OP_MAX_EVT_NAME_LEN - 1);
 		event.count = atoi(strtok(NULL, ":"));
 		/* Name and count are required in the event spec in order for
 		 * 'ophelp --check-events' to pass.  But since unit mask and domain
@@ -1170,6 +1365,7 @@ static void _process_events_list(void)
 		event.evt_um = 0ULL;
 		event.no_kernel = 0;
 		event.no_user = 0;
+		event.throttled = false;
 		memset(event.um_name, '\0', OP_MAX_UM_NAME_LEN);
 		while ((info = strtok(NULL, ":"))) {
 			switch (place) {
@@ -1179,7 +1375,7 @@ static void _process_events_list(void)
 				// consider the entire part a string.
 				if (*endptr) {
 					event.evt_um = 0;
-					strncpy(event.um_name, info, OP_MAX_UM_NAME_LEN);
+					strncpy(event.um_name, info, OP_MAX_UM_NAME_LEN - 1);
 				}
 				break;
 			case _OP_KERNEL:
@@ -1197,15 +1393,18 @@ static void _process_events_list(void)
 		_get_event_code(&event);
 		events.push_back(event);
 	}
-#if (defined(__powerpc__) || defined(__powerpc64__))
+#if PPC64_ARCH
 	{
-		/* This section of code is for architectures such as ppc[64] for which
-		 * the oprofile event code needs to be converted to the appropriate event
-		 * code to pass to the perf_event_open syscall.
+		/* For ppc64 architecture processors prior to the introduction of
+		 * architected_events_v1, the oprofile event code needs to be converted
+		 * to the appropriate event code to pass to the perf_event_open syscall.
+		 * But as of the introduction of architected_events_v1, the events
+		 * file contains the necessary event code information, so this conversion
+		 * step is no longer needed.
 		 */
 
 		using namespace OP_perf_utils;
-		if (!op_convert_event_vals(&events)) {
+		if ((cpu_type < CPU_PPC64_ARCH_V1) && !op_convert_event_vals(&events)) {
 			cerr << "Unable to convert all oprofile event values to perf_event values" << endl;
 			exit(EXIT_FAILURE);
 		}
@@ -1244,7 +1443,7 @@ static void get_default_event(void)
 	_get_event_code(&dft_evt);
 	events.push_back(dft_evt);
 
-#if (defined(__powerpc__) || defined(__powerpc64__))
+#if PPC64_ARCH
 	{
 		/* This section of code is for architectures such as ppc[64] for which
 		 * the oprofile event code needs to be converted to the appropriate event
@@ -1252,7 +1451,7 @@ static void get_default_event(void)
 		 */
 
 		using namespace OP_perf_utils;
-		if (!op_convert_event_vals(&events)) {
+		if ((cpu_type < CPU_PPC64_ARCH_V1) && !op_convert_event_vals(&events)) {
 			cerr << "Unable to convert all oprofile event values to perf_event values" << endl;
 			exit(EXIT_FAILURE);
 		}
@@ -1268,6 +1467,10 @@ static void _process_session_dir(void)
 		cwd = (char *) xmalloc(PATH_MAX);
 		// set default session dir
 		cwd = getcwd(cwd, PATH_MAX);
+		if (cwd == NULL) {
+			perror("Error calling getcwd");
+			exit(EXIT_FAILURE);
+		}
 		operf_options::session_dir = cwd;
 		operf_options::session_dir +="/oprofile_data";
 		samples_dir = operf_options::session_dir + "/samples";
@@ -1304,6 +1507,7 @@ static void _process_session_dir(void)
 			exit(EXIT_FAILURE);
 		}
 	}
+
 	cverb << vdebug << "Using samples dir " << samples_dir << endl;
 }
 
@@ -1535,21 +1739,18 @@ static void process_args(int argc, char * const argv[])
 
 		app_name = (char *) xmalloc(strlen(argv[non_options_idx]) + 1);
 		strcpy(app_name, argv[non_options_idx]);
+		// Note 1: app_args[0] is placeholder for app_fname (filled in later).
+		// Note 2: app_args[<end>] is set to NULL (required by execvp)
 		if (non_options_idx < (argc -1)) {
-			size_t length_all_app_args = 0;
-			non_options_idx++;
-			for(int i = non_options_idx; i < argc; i++)
-				length_all_app_args += strlen(argv[i]) + 1;
-			if (length_all_app_args) {
-				app_args = (char *) xmalloc(length_all_app_args);
-				strcpy(app_args, argv[non_options_idx]);
-				non_options_idx++;
-				while (non_options_idx < argc) {
-					strcat(app_args, " ");
-					strcat(app_args, argv[non_options_idx]);
-					non_options_idx++;
-				}
+			app_args = (char **) xmalloc((sizeof *app_args) *
+			                             (argc - non_options_idx + 1));
+			for(int i = non_options_idx + 1; i < argc; i++) {
+				app_args[i - non_options_idx] = argv[i];
 			}
+			app_args[argc - non_options_idx] = NULL;
+		} else {
+			app_args = (char **) xmalloc((sizeof *app_args) * 2);
+			app_args[1] = NULL;
 		}
 		if (validate_app_name() < 0) {
 			__print_usage_and_exit(NULL);
@@ -1586,7 +1787,7 @@ static void process_args(int argc, char * const argv[])
 	} else  {
 		_process_events_list();
 	}
-	op_nr_counters = events.size();
+	op_nr_events = events.size();
 
 	if (operf_options::vmlinux.empty()) {
 		no_vmlinux = true;
@@ -1599,7 +1800,67 @@ static void process_args(int argc, char * const argv[])
 	return;
 }
 
-static int _check_perf_events_cap(void)
+static int _get_cpu_for_perf_events_cap(void)
+{
+	int retval;
+	string err_msg;
+	char cpus_online[257];
+	FILE * online_cpus;
+	DIR *dir = NULL;
+
+	int total_cpus = sysconf(_SC_NPROCESSORS_ONLN);
+	if (!total_cpus) {
+		err_msg = "Internal Error (1): Number of online cpus cannot be determined.";
+		retval = -1;
+		goto error;
+	}
+
+	online_cpus = fopen("/sys/devices/system/cpu/online", "r");
+	if (!online_cpus) {
+		err_msg = "Internal Error (2): Number of online cpus cannot be determined.";
+		retval = -1;
+		goto error;
+	}
+	memset(cpus_online, 0, sizeof(cpus_online));
+
+	if ( fgets(cpus_online, sizeof(cpus_online), online_cpus) == NULL) {
+		fclose(online_cpus);
+		err_msg = "Internal Error (3): Number of online cpus cannot be determined.";
+		retval = -1;
+		goto error;
+	}
+
+	if (!cpus_online[0]) {
+		fclose(online_cpus);
+		err_msg = "Internal Error (4): Number of online cpus cannot be determined.";
+		retval = -1;
+		goto error;
+
+	}
+	if (index(cpus_online, ',') || cpus_online[0] != '0') {
+		// A comma in cpus_online implies a gap, which in turn implies that not all
+		// CPUs are online.
+		if ((dir = opendir("/sys/devices/system/cpu")) == NULL) {
+			fclose(online_cpus);
+			err_msg = "Internal Error (5): Number of online cpus cannot be determined.";
+			retval = -1;
+			goto error;
+		} else {
+			struct dirent *entry = NULL;
+			retval = OP_perf_utils::op_get_next_online_cpu(dir, entry);
+			closedir(dir);
+		}
+	} else {
+		// All CPUs are available, so we just arbitrarily choose CPU 0.
+		retval = 0;
+	}
+	fclose(online_cpus);
+error:
+	return retval;
+}
+
+
+static int _check_perf_events_cap(bool use_cpu_minus_one)
 {
 	/* If perf_events syscall is not implemented, the syscall below will fail
 	 * with ENOSYS (38).  If implemented, but the processor type on which this
@@ -1608,12 +1869,14 @@ static int _check_perf_events_cap(void)
 	 */
 	struct perf_event_attr attr;
 	pid_t pid ;
+	int cpu_to_try = use_cpu_minus_one ? -1 : _get_cpu_for_perf_events_cap();
+	errno = 0;
         memset(&attr, 0, sizeof(attr));
         attr.size = sizeof(attr);
         attr.sample_type = PERF_SAMPLE_IP;
 
 	pid = getpid();
-	syscall(__NR_perf_event_open, &attr, pid, 0, -1, 0);
+	syscall(__NR_perf_event_open, &attr, pid, cpu_to_try, -1, 0);
 	return errno;
 
 }
@@ -1665,21 +1928,43 @@ static int _get_sys_value(const char * filename)
 int main(int argc, char * const argv[])
 {
 	int rc;
+	int perf_event_paranoid = _get_sys_value("/proc/sys/kernel/perf_event_paranoid");
+
+	my_uid = geteuid();
 	throttled = false;
-	if ((rc = _check_perf_events_cap())) {
-		if (rc == EBUSY) {
-			cerr << "Performance monitor unit is busy.  Do 'opcontrol --deinit' and try again." << endl;
-			exit(1);
-		}
-		if (rc == ENOSYS) {
-			cerr << "Your kernel does not implement a required syscall"
-			     << "  for the operf program." << endl;
-		} else if (rc == ENOENT) {
-			cerr << "Your kernel's Performance Events Subsystem does not support"
-			     << " your processor type." << endl;
-		} else {
-			cerr << "Unexpected error running operf: " << strerror(rc) << endl;
+	rc = _check_perf_events_cap(use_cpu_minus_one);
+	if (rc == EACCES) {
+		/* Early perf_events kernels required the cpu argument to perf_event_open
+		 * to be '-1' when setting up to profile a single process if 1) the user is
+		 * not root; and 2) perf_event_paranoid is > 0.  An EACCES error would be
+		 * returned if passing '0' or greater for the cpu arg and the above criteria
+		 * was not met.  Unfortunately, later kernels turned this requirement around
+		 * such that the passed cpu arg must be '0' or greater when the user is not
+		 * root.
+		 *
+		 * We don't really have a good way to check whether we're running on such an
+		 * early kernel except to try the perf_event_open with different values to see
+		 * what works.
+		 */
+		if (my_uid != 0 && perf_event_paranoid > 0) {
+			use_cpu_minus_one = true;
+			rc = _check_perf_events_cap(use_cpu_minus_one);
 		}
+	}
+	if (rc == EBUSY) {
+		cerr << "Performance monitor unit is busy.  Do 'opcontrol --deinit' and try again." << endl;
+		exit(1);
+	}
+	if (rc == ENOSYS) {
+		cerr << "Your kernel does not implement a required syscall"
+		     << " for the operf program." << endl;
+	} else if (rc == ENOENT) {
+		cerr << "Your kernel's Performance Events Subsystem does not support"
+		     << " your processor type." << endl;
+	} else if (rc) {
+		cerr << "Unexpected error running operf: " << strerror(rc) << endl;
+	}
+	if (rc) {
 		cerr << "Please use the opcontrol command instead of operf." << endl;
 		exit(1);
 	}
@@ -1688,8 +1973,6 @@ int main(int argc, char * const argv[])
 	cpu_speed = op_cpu_frequency();
 	process_args(argc, argv);
 
-	int perf_event_paranoid = _get_sys_value("/proc/sys/kernel/perf_event_paranoid");
-	my_uid = geteuid();
 	if (operf_options::system_wide && ((my_uid != 0) && (perf_event_paranoid > 0))) {
 		cerr << "To do system-wide profiling, either you must be root or" << endl;
 		cerr << "/proc/sys/kernel/perf_event_paranoid must be set to 0 or -1." << endl;
