@@ -9,10 +9,14 @@
  * @author Philippe Elie
  */
 
+#include <cstdlib>
+
 #include <iostream>
 #include <sstream>
 #include <iterator>
+#include <cstdlib>
 
+#include "op_config.h"
 #include "locate_images.h"
 #include "op_exception.h"
 #include "popt_options.h"
@@ -20,27 +24,40 @@
 #include "common_option.h"
 #include "file_manip.h"
 
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <limits.h>
+
 using namespace std;
 
 namespace options {
-	extra_images extra_found_images;
 	double threshold = 0.0;
 	string threshold_opt;
+	string session_dir;
+	string command_options;
+	vector<string> image_path;
+	string root_path;
 }
 
 namespace {
 
-vector<string> image_path;
+
 vector<string> verbose_strings;
 
 popt::option common_options_array[] = {
 	popt::option(verbose_strings, "verbose", 'V',
 		     // FIXME help string for verbose level
-		     "verbose output", "all,debug,bfd,level1,sfile,stats"),
-	popt::option(image_path, "image-path", 'p',
+		     "verbose output", "all,debug,bfd,level1,sfile,stats,xml"),
+	popt::option(options::session_dir, "session-dir", '\0',
+		     "specify session path to hold samples database and session data (" OP_SESSION_DIR_DEFAULT ")", "path"),
+	popt::option(options::image_path, "image-path", 'p',
 		     "comma-separated path to search missing binaries", "path"),
+	popt::option(options::root_path, "root", 'R',
+		     "path to filesystem to search for missing binaries", "path"),
 };
 
+int session_dir_supplied;
 
 double handle_threshold(string threshold)
 {
@@ -153,12 +170,31 @@ fail:
 	exit(EXIT_FAILURE);
 }
 
-
 options::spec get_options(int argc, char const * argv[])
 {
 	vector<string> non_options;
 	popt::parse_options(argc, argv, non_options);
 
+	// initialize paths in op_config.h
+	if (options::session_dir.empty()) {
+		char * cwd;
+		struct stat sb;
+		// If <curr_dir>/oprofile_data exists (used by operf), we'll use that as session-dir
+		cwd = new char[PATH_MAX];
+		options::session_dir = (getcwd(cwd, PATH_MAX) == NULL) ? "" : cwd;
+		delete [] cwd;
+		options::session_dir +="/oprofile_data";
+		if ((stat(options::session_dir.c_str(), &sb) < 0) ||
+				((sb.st_mode & S_IFMT) != S_IFDIR)) {
+			// Use the standard default session dir instead
+			options::session_dir = "/var/lib/oprofile";
+		}
+		session_dir_supplied = 0;
+	} else {
+		session_dir_supplied = 1;
+	}
+	init_op_config_dirs(options::session_dir.c_str());
+
 	if (!options::threshold_opt.empty())
 		options::threshold = handle_threshold(options::threshold_opt);
 
@@ -167,19 +203,11 @@ options::spec get_options(int argc, char const * argv[])
 		exit(EXIT_FAILURE);
 	}
 
-	bool ok = true;
-	vector<string>::const_iterator it;
-	for (it = image_path.begin(); it != image_path.end(); ++it) {
-		if (!is_directory(*it)) {
-			cerr << *it << " isn't a valid directory\n";
-			ok = false;
-		}
-	}
-
-	if (!ok)
-		throw op_runtime_error("invalid --image-path= options");
-
-	options::extra_found_images.populate(image_path);
+	// XML generator needs command line options for its header
+	ostringstream str;
+	for (int i = 1; i < argc; ++i)
+		str << argv[i] << " ";
+	options::command_options = str.str();
 
 	return parse_spec(non_options);
 }
@@ -285,3 +313,8 @@ merge_option handle_merge_option(vector<string> const & mergespec,
 
 	return merge_by;
 }
+
+bool was_session_dir_supplied(void)
+{
+	return session_dir_supplied;
+}
